!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Mathematical and Computational Sciences Division                !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module patch_problem

!----------------------------------------------------
! This module contains routines to create, solve and analyze local problems
! on a patch of elements.
!
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global, only : my_real, ierr, colon, outunit, global_element_kind, &
                   ALLOC_FAILED, PHAML_INTERNAL_ERROR, DIRICHLET, NATURAL, &
                   INTERIOR, MINIMUM_RULE, MAXIMUM_RULE, TETRAHEDRAL_ELEMENT
use message_passing, only : fatal
use gridtype_mod, only : grid_type, BOUNDARY, BASIS_DIM, &
                         VERTICES_PER_ELEMENT, EDGES_PER_ELEMENT, &
                         NEIGHBORS_PER_ELEMENT, &
                         size_elem_soln, get_elem_soln, size_edge_soln, &
                         get_edge_soln, size_vert_soln, get_vert_soln
use grid_util, only : get_neighbors, get_vertex_elements, element_dof, my_bconds
use hash_mod, only : HASH_NOT_FOUND, hash_decode_key, operator(==), &
                     operator(/)
use search_tree_mod, only : search_tree_type, create_search_tree, &
                            destroy_search_tree, insert_search_tree, &
                            id_is_in_search_tree, get_search_tree_data, &
                            ST_NO_DATA
use make_linsys, only : elemental_matrix
use linsystype_mod, only : solver_options
use quadrature_rules, only : quadrature_rule_line, MAX_QUAD_ORDER_LINE
use basis_functions, only : p_hier_basis_func
use evaluate, only : evaluate_soln_local

!----------------------------------------------------

implicit none
private

! All routines are thread safe.

public patch_type, create_patch, destroy_patch, reset_patch, init_patch, &
       h_refine_patch_element, h_refine_patch, p_refine_patch_element, &
       p_refine_patch, set_p_patch, get_patch_number_of_elements, &
       create_patch_linear_system, destroy_patch_linear_system, &
       solve_patch_linear_system, get_patch_linear_system_dof, &
       norm_patch_diff, norm_patch_soln, print_patch_grid, &
       print_patch_linear_system, LOCAL_H_ERRIND_PATCH

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

integer, parameter :: DEFAULT_PATCH_TYPE   = 0, &
                      LOCAL_H_ERRIND_PATCH = 1

!----------------------------------------------------
! The following types are defined:

type point_type
   real(my_real) :: x, y ! Coordinates
end type point_type

type vertex_type
   type(point_type) :: coord                       ! The coordinates
   integer :: bmark                                ! Boundary condition marker
   integer, pointer :: type(:)                     ! Vertex type (eg Dirichlet)
   integer :: grid_id                              ! Index of vertex in grid or
                                                   ! 0 if not in grid
   integer :: first_eq                             ! The index of the first
                                                   ! linear system equation
                                                   ! associated with this vertex
   integer :: neq                                  ! Number of equations
                                                   ! associated with the vertex
   real(my_real), pointer :: solution(:,:)=>NULL() ! Solution at this vertex.
                                                   ! (comp,eigen)
   real(my_real), pointer :: grid_solution(:,:) &  ! Solution copied from
                             =>NULL()              ! the grid
! NOTE: For vertices created by h-refinement of patch elements, grid_solution
!       is set to the linear interpolation of the grid_solution of the
!       end points of the refinement edge, which may not agree with the
!       grid's solution if that vertex exists in the grid.
end type vertex_type

type edge_type
   integer :: vertex(2)                            ! The vertices
   integer :: degree                               ! Polynomial degree
   integer :: bmark                                ! Boundary condition marker
   integer, pointer :: type(:)                     ! Edge type (Dirichlet, etc.)
   logical :: isleaf                               ! Has it been refined?
   integer :: grid_id                              ! Index of edge in grid or
                                                   ! 0 if not in grid
   integer :: first_eq                             ! The index of the first
                                                   ! linear system equation
                                                   ! associated with this edge
   integer :: neq                                  ! Number of equations
                                                   ! associated with this edge
   real(my_real), pointer :: solution(:,:,:) &     ! Solution coefficients at
                             =>NULL()              ! this edge.
                                                   ! (basis,comp,eigen)
   real(my_real), pointer :: grid_solution(:,:,:) &! Solution copied from
                             =>NULL()              ! the grid
! NOTE: For edges created by h-refinement of patch elements, grid_solution
!       is set to 0, which may not agree with the grid's solution if that
!       edge exists in the grid.
end type edge_type

type element_type
   integer :: vertex(VERTICES_PER_ELEMENT)         ! The vertices
   integer :: edge(EDGES_PER_ELEMENT)              ! The edges.  i'th edge is
                                                   ! opposite i'th vertex.
   integer :: neighbor(NEIGHBORS_PER_ELEMENT)      ! Neighboring elements.  i'th
                                                   ! neighbor is opposite i'th
                                                   ! vertex.  3rd neighbor is
                                                   ! mate or parent of mate.
   integer :: degree                               ! Polynomial degree
   integer :: level                                ! h-refinement level
   logical :: isleaf                               ! Is it a leaf element?
   integer :: grid_id                              ! Index of element in grid or
                                                   ! the ancestor element in
                                                   ! grid if not in grid
   integer :: first_eq                             ! The index of the first
                                                   ! linear system equation
                                                   ! associated with this elem
   integer :: neq                                  ! Number of equations
                                                   ! associated with this elem
   real(my_real), pointer :: solution(:,:,:) &     ! Solution coefficients for
                             =>NULL()              ! this element.
                                                   ! (basis,comp,eigen)
   real(my_real), pointer :: grid_solution(:,:,:) &! Solution copied from
                             =>NULL()              ! the grid
! NOTE: For elements created by h-refinement of patch elements, grid_solution
!       is set to 0, which may not agree with the grid's solution if that
!       element exists in the grid.
end type element_type

type patch_type
   private

! The patch grid

   type(vertex_type), pointer :: vertex(:)=>NULL()   ! The vertices
   type(edge_type), pointer :: edge(:)=>NULL()       ! The edges
   type(element_type), pointer :: element(:)=>NULL() ! The elements
   integer :: nvert, nedge, nelem                    ! Number of patch entities
   integer :: nlev                                   ! Number of h-refinement
                                                     ! levels
   integer :: system_size                            ! Multicomponent solutions
   integer :: num_evec                               ! Number of eigenvectors
                                                     ! 1 if not an eigenvalue
                                                     ! problem
   integer :: type_of_patch                          ! Designate special patches

! The patch linear system

   real(my_real), pointer :: A(:,:)=>NULL()  ! The linear system matrix
   real(my_real), pointer :: LU(:,:)=>NULL() ! The factored matrix
   real(my_real), pointer :: b(:,:)=>NULL()  ! The right hand side. Second
                                             ! dimension is multiple
                                             ! eigenvectors
   real(my_real), pointer :: x(:,:)=>NULL()  ! The linear system solution
   integer, pointer :: ipiv(:)=>NULL()       ! The pivot array for LU
                                             ! factorization
   integer :: neq                            ! Size of full linear system
   integer :: n                              ! Size of system with Dirichlet
                                             ! rows removed
   logical, pointer :: dirichlet(:)=>NULL()  ! Indicates which rows are
                                             ! Dirichlet
   integer, pointer :: nondir_rows(:)=>NULL()! A vector of size n giving the
                                             ! indices of the non-Dirichlet rows
                                             ! for correspondence between A and
                                             ! LU, and b and x
   integer, pointer :: dir_rows(:)=>NULL()   ! Vector of Dirichlet rows
end type patch_type

!----------------------------------------------------
! The following variables are defined:

!----------------------------------------------------

contains

!====================================================
! Public routines
!====================================================

!          ------------
subroutine create_patch(patch)
!          ------------

!----------------------------------------------------
! This subroutine creates the data structure for a patch with an initial patch
! size of 16 triangles, which would be enough for a 1-level halo with vertex
! neighbors in an isosceles right triangle grid.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(out) :: patch ! The patch to create
!----------------------------------------------------
! Local variables:

integer :: astat
!----------------------------------------------------
! Begin executable code

! Allocate the patch grid entities

allocate(patch%vertex(16),patch%edge(32),patch%element(16),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_patch")
   stop
endif

! There are no grid entities yet

patch%nvert = 0
patch%nedge = 0
patch%nelem = 0

! There is no linear system yet

patch%neq = 0

! Assume it is a normal patch

patch%type_of_patch = DEFAULT_PATCH_TYPE

end subroutine create_patch

!          -------------
subroutine destroy_patch(patch)
!          -------------

!----------------------------------------------------
! This subroutine destroys the patch, freeing all allocated memory.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch to destroy
!----------------------------------------------------
! Local variables:

integer :: i
!----------------------------------------------------
! Begin executable code

! Destroy the patch linear system

call destroy_patch_linear_system(patch)

! Deallocate all grid memory

do i=1,patch%nvert
   if (associated(patch%vertex(i)%type)) deallocate(patch%vertex(i)%type)
   if (associated(patch%vertex(i)%solution))deallocate(patch%vertex(i)%solution)
   if (associated(patch%vertex(i)%grid_solution)) &
      deallocate(patch%vertex(i)%grid_solution)
end do

do i=1,patch%nedge
   if (associated(patch%edge(i)%type)) deallocate(patch%edge(i)%type)
   if (associated(patch%edge(i)%solution)) deallocate(patch%edge(i)%solution)
   if (associated(patch%edge(i)%grid_solution)) &
      deallocate(patch%edge(i)%grid_solution)
end do

do i=1,patch%nelem
   if (associated(patch%element(i)%solution)) &
      deallocate(patch%element(i)%solution)
   if (associated(patch%element(i)%grid_solution)) &
      deallocate(patch%element(i)%grid_solution)
end do

if (associated(patch%vertex)) deallocate(patch%vertex)
if (associated(patch%edge)) deallocate(patch%edge)
if (associated(patch%element)) deallocate(patch%element)

end subroutine destroy_patch

!          -----------
subroutine reset_patch(patch)
!          -----------

!----------------------------------------------------
! This subroutine invalidates a patch, essentially returning it to the state
! a patch is in after create_patch, although the allocated size of the patch
! may be larger than what create_patch creates.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch to reset
!----------------------------------------------------
! Local variables:

integer :: i
!----------------------------------------------------
! Begin executable code

! Destroy the patch linear system

call destroy_patch_linear_system(patch)

! Deallocate memory within each grid entity

do i=1,patch%nvert
   if (associated(patch%vertex(i)%type)) deallocate(patch%vertex(i)%type)
   if (associated(patch%vertex(i)%solution))deallocate(patch%vertex(i)%solution)
   if (associated(patch%vertex(i)%grid_solution)) &
      deallocate(patch%vertex(i)%grid_solution)
end do

do i=1,patch%nedge
   if (associated(patch%edge(i)%type)) deallocate(patch%edge(i)%type)
   if (associated(patch%edge(i)%solution)) deallocate(patch%edge(i)%solution)
   if (associated(patch%edge(i)%grid_solution)) &
      deallocate(patch%edge(i)%grid_solution)
end do

do i=1,patch%nelem
   if (associated(patch%element(i)%solution)) &
      deallocate(patch%element(i)%solution)
   if (associated(patch%element(i)%grid_solution)) &
      deallocate(patch%element(i)%grid_solution)
end do

! There are no grid entities any more

patch%nvert = 0
patch%nedge = 0
patch%nelem = 0

! Reset to a normal patch

patch%type_of_patch = DEFAULT_PATCH_TYPE

end subroutine reset_patch

!          ----------
subroutine init_patch(grid,patch,core,include_mate,halo_size, &
                      include_edge_neigh,include_vert_neigh,boundary_condition,&
                      eigen,special)
!          ----------

!----------------------------------------------------
! This subroutine initializes a patch by copying the data for element core and
! the designated neighbors from grid.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid    ! The grid the patch is extracted from
type(patch_type), intent(inout) :: patch  ! The patch to initialize
integer, intent(in) :: core               ! The element at the core of the patch
logical, intent(in) :: include_mate       ! If true, also include core's mate in
                                          ! the core of the patch
integer, intent(in) :: halo_size          ! How many layers around the core
logical, intent(in) :: include_edge_neigh ! If true, also include neighbors
                                          ! that only share an edge.
                                          ! Irrelevant in 2D.
logical, intent(in) :: include_vert_neigh ! If true, also include neighbors
                                          ! that only share a vertex.
integer, intent(in) :: boundary_condition ! How to set vertex type, edge type,
                                          ! and bmark for the patch boundary:
                                          ! 0 - keep the values from grid
                                          ! 1 - set all Dirichlet with
                                          !     bmark = huge(0)
                                          ! 2 - set grid interior Dirichlet with
                                          !     bmark = huge(0)
                                          ! 3 - set all Natural with 
                                          !     bmark = huge(0)
                                          ! 4 - set grid interior Natural with
                                          !     bmark = huge(0)
                                          ! 5 - set all vertices Dirichlet with
                                          !     bmark = huge(0) and interior
                                          !     edges Dirichlet bmark = huge(0)
                                          ! 6 - set all vertices Dirichlet with
                                          !     bmark = huge(0) and interior
                                          !     edges and high order edges
                                          !     Dirichlet with bmark = huge(0)
integer, intent(in), optional :: eigen    ! If present, which eigen to use for
                                          ! solution.  Otherwise do all.
integer, intent(in), optional :: special  ! Patches with special handling, e.g.
                                          ! local h error indicator.
!----------------------------------------------------
! Local variables:

type(search_tree_type) :: patch_vert_id, patch_edge_id, patch_elem_id
integer :: e, v, elem, edge, vert, patch_elem, grid_elem, patch_edge, &
           grid_edge, patch_vert, grid_vert, mate, halo, start, &
           last_before_halo, minlev, maxlev, nsoln, astat, jerr, dim2, dim3
integer, pointer :: neighbor(:)
!----------------------------------------------------
! Begin executable code

! TEMP Check for things that have not been implemented yet.

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("init_patch: patch problems not yet implemented for tetrahedra")
   stop
endif

if (grid%any_periodic) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("init_patch: patch problems not yet implemented for periodic boundary conditions")
   stop
endif

! The patch should not already exist

if (patch%nelem > 0) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("init_patch: attempt to initialize an already existing patch")
   stop
endif

! But the data structure should already have been created

if (.not. associated(patch%element)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("init_patch: attempt to initialize a patch that has not been created")
   stop
endif

! Type of patch; it was set to default in create and reset

if (present(special)) then
   patch%type_of_patch = special
endif

! Number of eigenvectors and components

if (present(eigen)) then
   patch%num_evec = 1
else
   patch%num_evec = max(1,grid%num_eval)
endif
patch%system_size = grid%system_size

! Create the correspondence between elements in the patch and elements in grid

call create_search_tree(patch_elem_id)

! The core element

patch%nelem = 1
patch%element(1)%grid_id = core
call insert_search_tree(patch_elem_id,core,1,jerr)

! The mate, if requested and it exists.  By the mate we mean the third
! neighbor even if it is the parent of the actual mate.

if (include_mate) then
   if (.not. grid%element(core)%mate == BOUNDARY) then
      mate = hash_decode_key(grid%element(core)%mate,grid%elem_hash)
      if (mate == HASH_NOT_FOUND) then
         mate = hash_decode_key(grid%element(core)%mate/2,grid%elem_hash)
      endif
      if (mate == HASH_NOT_FOUND) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("init_patch: mate does not exist")
         stop
      endif
      patch%nelem = 2
      patch%element(2)%grid_id = mate
      call insert_search_tree(patch_elem_id,mate,2,jerr)
   endif
endif

! For each halo level, get the neighbors and add them to the patch

start = 1
last_before_halo = patch%nelem
do halo=1,halo_size
   do patch_elem=start,last_before_halo
      grid_elem = patch%element(patch_elem)%grid_id
      do v=1,VERTICES_PER_ELEMENT
         if (include_vert_neigh) then
            vert = grid%element(grid_elem)%vertex(v)
            call get_vertex_elements(grid,vert,neighbor,include_periodic=.true.)
         else
            allocate(neighbor(NEIGHBORS_PER_ELEMENT),stat=astat)
            if (astat /= 0) then
               ierr = ALLOC_FAILED
               call fatal("memory allocation failed in init_patch")
               stop
            endif
            neighbor =  get_neighbors(grid_elem,grid)
         endif
         do e=1,size(neighbor)
            if (neighbor(e) == BOUNDARY) cycle
            if (.not. id_is_in_search_tree(patch_elem_id,neighbor(e))) then
               patch%nelem = patch%nelem + 1
               if (patch%nelem > size(patch%element)) then
                  call more_elements(patch)
               endif
               patch%element(patch%nelem)%grid_id = neighbor(e)
               call insert_search_tree(patch_elem_id,neighbor(e),patch%nelem, &
                                       jerr)
            endif
         end do
         deallocate(neighbor)
         if (.not. include_vert_neigh) exit
      end do
   end do
   start = last_before_halo + 1
   last_before_halo = patch%nelem
end do

! Create the correspondence between edges in the patch and edges in grid

call create_search_tree(patch_edge_id)
patch%nedge = 0
do elem=1,patch%nelem
   do e=1,EDGES_PER_ELEMENT
      edge = grid%element(patch%element(elem)%grid_id)%edge(e)
      if (.not. id_is_in_search_tree(patch_edge_id,edge)) then
         patch%nedge = patch%nedge + 1
         if (patch%nedge > size(patch%edge)) then
            call more_edges(patch)
         endif
         patch%edge(patch%nedge)%grid_id = edge
         call insert_search_tree(patch_edge_id,edge,patch%nedge,jerr)
      endif
   end do
end do

! Create the correspondence between vertices in the patch and vertices in grid

call create_search_tree(patch_vert_id)
patch%nvert = 0
do elem=1,patch%nelem
   do v=1,VERTICES_PER_ELEMENT
      vert = grid%element(patch%element(elem)%grid_id)%vertex(v)
      if (.not. id_is_in_search_tree(patch_vert_id,vert)) then
         patch%nvert = patch%nvert + 1
         if (patch%nvert > size(patch%vertex)) then
            call more_vertices(patch)
         endif
         patch%vertex(patch%nvert)%grid_id = vert
         call insert_search_tree(patch_vert_id,vert,patch%nvert,jerr)
      endif
   end do
end do

! Copy the elements

allocate(neighbor(NEIGHBORS_PER_ELEMENT),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in init_patch")
   stop
endif
minlev = huge(0)
maxlev = -huge(0)
do patch_elem=1,patch%nelem
   grid_elem = patch%element(patch_elem)%grid_id
   do v=1,VERTICES_PER_ELEMENT
      grid_vert = grid%element(grid_elem)%vertex(v)
      patch%element(patch_elem)%vertex(v) = &
         get_search_tree_data(patch_vert_id,grid_vert)
   end do
   do e=1,EDGES_PER_ELEMENT
      grid_edge = grid%element(grid_elem)%edge(e)
      patch%element(patch_elem)%edge(e) = &
         get_search_tree_data(patch_edge_id,grid_edge)
   end do
   neighbor =  get_neighbors(grid_elem,grid)
   do e=1,NEIGHBORS_PER_ELEMENT
      if (neighbor(e) == BOUNDARY) then
         patch%element(patch_elem)%neighbor(e) = BOUNDARY
      else
         patch%element(patch_elem)%neighbor(e) = &
            get_search_tree_data(patch_elem_id,neighbor(e))
         if (patch%element(patch_elem)%neighbor(e) == ST_NO_DATA) then
            patch%element(patch_elem)%neighbor(e) = BOUNDARY
         endif
      endif
   end do
   patch%element(patch_elem)%degree = grid%element(grid_elem)%degree

   patch%element(patch_elem)%level = grid%element(grid_elem)%level
   minlev = min(minlev,patch%element(patch_elem)%level)
   maxlev = max(maxlev,patch%element(patch_elem)%level)
   nsoln = size_elem_soln(grid,grid_elem,BASIS_DIM)
   if (nsoln /= 0) then
      dim2 = patch%system_size
      dim3 = patch%num_evec
      allocate(patch%element(patch_elem)%grid_solution(nsoln,dim2,dim3), &
               patch%element(patch_elem)%solution(nsoln,dim2,dim3), stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in init_patch")
         stop
      endif
      if (present(eigen)) then
         patch%element(patch_elem)%grid_solution(:,:,1) = get_elem_soln(grid, &
                                                    grid_elem,colon,colon,eigen)
      else
         patch%element(patch_elem)%grid_solution = get_elem_soln(grid, &
                                                   grid_elem,colon,colon,colon)
      endif
      patch%element(patch_elem)%solution = 0.0_my_real
   endif
end do
deallocate(neighbor)

! Adjust the levels so the minimum level is 1

patch%element(1:patch%nelem)%level = &
   patch%element(1:patch%nelem)%level - minlev + 1
patch%nlev = maxlev - minlev + 1

! All elements and edges in the initial patch are leaves

patch%element(1:patch%nelem)%isleaf = .true.
patch%edge(1:patch%nedge)%isleaf = .true.

! Copy the edges

do patch_edge=1,patch%nedge
   grid_edge = patch%edge(patch_edge)%grid_id
   patch%edge(patch_edge)%vertex(1) = &
      get_search_tree_data(patch_vert_id,grid%edge(grid_edge)%vertex(1))
   patch%edge(patch_edge)%vertex(2) = &
      get_search_tree_data(patch_vert_id,grid%edge(grid_edge)%vertex(2))
   patch%edge(patch_edge)%degree = grid%edge(grid_edge)%degree
   patch%edge(patch_edge)%bmark = grid%edge(grid_edge)%bmark
   allocate(patch%edge(patch_edge)%type(grid%system_size),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in init_patch")
      stop
   endif
   patch%edge(patch_edge)%type = grid%edge_type(grid_edge,:)
   nsoln = size_edge_soln(grid,grid_edge,BASIS_DIM)
   if (nsoln /= 0) then
      dim2 = patch%system_size
      dim3 = patch%num_evec
      allocate(patch%edge(patch_edge)%grid_solution(nsoln,dim2,dim3), &
               patch%edge(patch_edge)%solution(nsoln,dim2,dim3), stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in init_patch")
         stop
      endif
      if (present(eigen)) then
         patch%edge(patch_edge)%grid_solution(:,:,1) = get_edge_soln(grid, &
                                                    grid_edge,colon,colon,eigen)
      else
         patch%edge(patch_edge)%grid_solution = get_edge_soln(grid,grid_edge, &
                                                              colon,colon,colon)
      endif
      patch%edge(patch_edge)%solution = 0.0_my_real
   endif
end do

! Copy the vertices

do patch_vert=1,patch%nvert
   grid_vert = patch%vertex(patch_vert)%grid_id
   patch%vertex(patch_vert)%coord%x = grid%vertex(grid_vert)%coord%x
   patch%vertex(patch_vert)%coord%y = grid%vertex(grid_vert)%coord%y
   patch%vertex(patch_vert)%bmark = grid%vertex(grid_vert)%bmark
   allocate(patch%vertex(patch_vert)%type(grid%system_size),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in init_patch")
      stop
   endif
   patch%vertex(patch_vert)%type = grid%vertex_type(grid_vert,:)
   dim2 = patch%system_size
   dim3 = patch%num_evec
   allocate(patch%vertex(patch_vert)%grid_solution(dim2,dim3), &
            patch%vertex(patch_vert)%solution(dim2,dim3), stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in init_patch")
      stop
   endif
   if (present(eigen)) then
      patch%vertex(patch_vert)%grid_solution(:,1) = get_vert_soln(grid, &
                                                    grid_vert,colon,eigen)
   else
      patch%vertex(patch_vert)%grid_solution = get_vert_soln(grid,grid_vert, &
                                                             colon,colon)
   endif
   patch%vertex(patch_vert)%solution = 0.0_my_real
end do

! Set the boundary conditions

call set_patch_boundary_conditions(patch,boundary_condition)

call destroy_search_tree(patch_vert_id)
call destroy_search_tree(patch_edge_id)
call destroy_search_tree(patch_elem_id)

end subroutine init_patch

!                    ----------------------
recursive subroutine h_refine_patch_element(patch,elem)
!                    ----------------------

!----------------------------------------------------
! This subroutine h-refines element elem and any others needed for compatibility
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: elem              ! The element to refine
!----------------------------------------------------
! Local variables:

integer :: mate, child1, child2, child3, child4, edge1, edge2, edge3, edge4, &
           vert1, deg, astat, paredge, i
!----------------------------------------------------
! Begin executable code

! Identify the mate, creating it if necessary

mate = patch%element(elem)%neighbor(3)
if (mate /= BOUNDARY) then
   if (patch%element(mate)%level == patch%element(elem)%level-1) then
      call h_refine_patch_element(patch,mate)
      mate = patch%element(elem)%neighbor(3)
   endif
   if (patch%element(mate)%level /= patch%element(elem)%level) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_patch_element: mate does not have the same level as elem")
      stop
   endif
endif

! Get the indices for the new elements, edges and vertex

child1 = patch%nelem + 1
child2 = patch%nelem + 2
child3 = patch%nelem + 3
child4 = patch%nelem + 4
edge1  = patch%nedge + 1
edge2  = patch%nedge + 2
edge3  = patch%nedge + 3
edge4  = patch%nedge + 4
vert1  = patch%nvert + 1

! Expand memory if necessary

if (child4 > size(patch%element)) call more_elements(patch)
if (edge4  > size(patch%edge)) call more_edges(patch)
if (vert1  > size(patch%vertex)) call more_vertices(patch)

! The parent, mate and parent's 3rd edge are no longer a leaves

patch%element(elem)%isleaf = .false.
if (mate /= BOUNDARY) patch%element(mate)%isleaf = .false.
patch%edge(patch%element(elem)%edge(3))%isleaf = .false.

! For convenience

deg = patch%element(elem)%degree

! First child

patch%element(child1)%vertex = (/patch%element(elem)%vertex(1), &
                                 patch%element(elem)%vertex(3), &
                                 vert1 /)
patch%element(child1)%edge = (/edge3, edge1, &
                               patch%element(elem)%edge(2) /)
if (mate == BOUNDARY) then
   patch%element(child1)%neighbor = (/child2, BOUNDARY, &
                                      patch%element(elem)%neighbor(2) /)
else
   patch%element(child1)%neighbor = (/child2, child3, &
                                      patch%element(elem)%neighbor(2) /)
endif
do i=1,NEIGHBORS_PER_ELEMENT
  if (patch%element(elem)%neighbor(2) /= BOUNDARY) then
   if (patch%element(patch%element(elem)%neighbor(2))%neighbor(i) == elem) then
      patch%element(patch%element(elem)%neighbor(2))%neighbor(i) = child1
      exit
   endif
  endif
end do
patch%element(child1)%degree =  deg
patch%element(child1)%level = patch%element(elem)%level + 1
patch%element(child1)%isleaf = .true.
patch%element(child1)%grid_id = patch%element(elem)%grid_id
if (deg > 2) then
   allocate(patch%element(child1)%grid_solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
            patch%element(child1)%solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%element(child1)%grid_solution = 0.0_my_real
   patch%element(child1)%solution = 0.0_my_real
else
   nullify(patch%element(child1)%grid_solution, &
           patch%element(child1)%solution)
endif

! Second child

patch%element(child2)%vertex = (/patch%element(elem)%vertex(2), &
                                 patch%element(elem)%vertex(3), &
                                 vert1 /)
patch%element(child2)%edge = (/edge3, edge2, &
                               patch%element(elem)%edge(1) /)
if (mate == BOUNDARY) then
   patch%element(child2)%neighbor = (/child1, BOUNDARY, &
                                      patch%element(elem)%neighbor(1) /)
else
   patch%element(child2)%neighbor = (/child1, child4, &
                                      patch%element(elem)%neighbor(1) /)
endif
do i=1,NEIGHBORS_PER_ELEMENT
  if (patch%element(elem)%neighbor(1) /= BOUNDARY) then
   if (patch%element(patch%element(elem)%neighbor(1))%neighbor(i) == elem) then
      patch%element(patch%element(elem)%neighbor(1))%neighbor(i) = child2
      exit
   endif
  endif
end do
patch%element(child2)%degree =  deg
patch%element(child2)%level = patch%element(elem)%level + 1
patch%element(child2)%isleaf = .true.
patch%element(child2)%grid_id = patch%element(elem)%grid_id
if (deg > 2) then
   allocate(patch%element(child2)%grid_solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
            patch%element(child2)%solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%element(child2)%grid_solution = 0.0_my_real
   patch%element(child2)%solution = 0.0_my_real
else
   nullify(patch%element(child2)%grid_solution, &
           patch%element(child2)%solution)
endif

if (mate /= BOUNDARY) then

   deg = patch%element(mate)%degree

! Third child

   patch%element(child3)%vertex = (/patch%element(mate)%vertex(1), &
                                    patch%element(mate)%vertex(3), &
                                    vert1 /)
   patch%element(child3)%edge = (/edge4, edge1, &
                                  patch%element(mate)%edge(2) /)
   patch%element(child3)%neighbor = (/child4, child1, &
                                      patch%element(mate)%neighbor(2) /)
   do i=1,NEIGHBORS_PER_ELEMENT
     if (patch%element(mate)%neighbor(2) /= BOUNDARY) then
      if (patch%element(patch%element(mate)%neighbor(2))%neighbor(i)==mate) then
         patch%element(patch%element(mate)%neighbor(2))%neighbor(i) = child3
         exit
      endif
     endif
   end do
   patch%element(child3)%degree =  deg
   patch%element(child3)%level = patch%element(mate)%level + 1
   patch%element(child3)%isleaf = .true.
   patch%element(child3)%grid_id = patch%element(mate)%grid_id
   if (deg > 2) then
      allocate(patch%element(child3)%grid_solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
               patch%element(child3)%solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_patch_element")
         stop
      endif
      patch%element(child3)%grid_solution = 0.0_my_real
      patch%element(child3)%solution = 0.0_my_real
   else
      nullify(patch%element(child3)%grid_solution, &
              patch%element(child3)%solution)
   endif

! Fourth child

   patch%element(child4)%vertex = (/patch%element(mate)%vertex(2), &
                                    patch%element(mate)%vertex(3), &
                                    vert1 /)
   patch%element(child4)%edge = (/edge4, edge2, &
                                  patch%element(mate)%edge(1) /)
   patch%element(child4)%neighbor = (/child3, child2, &
                                      patch%element(mate)%neighbor(1) /)
   do i=1,NEIGHBORS_PER_ELEMENT
     if (patch%element(mate)%neighbor(1) /= BOUNDARY) then
      if (patch%element(patch%element(mate)%neighbor(1))%neighbor(i)==mate) then
         patch%element(patch%element(mate)%neighbor(1))%neighbor(i) = child4
         exit
      endif
     endif
   end do
   patch%element(child4)%degree =  deg
   patch%element(child4)%level = patch%element(mate)%level + 1
   patch%element(child4)%isleaf = .true.
   patch%element(child4)%grid_id = patch%element(mate)%grid_id
   if (deg > 2) then
      allocate(patch%element(child4)%grid_solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
               patch%element(child4)%solution(((deg-2)*(deg-1))/2, &
                                            patch%system_size,patch%num_evec), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_patch_element")
         stop
      endif
      patch%element(child4)%grid_solution = 0.0_my_real
      patch%element(child4)%solution = 0.0_my_real
   else
      nullify(patch%element(child4)%grid_solution, &
              patch%element(child4)%solution)
   endif

endif

! New vertex

patch%vertex(vert1)%coord%x = &
   (patch%vertex(patch%element(elem)%vertex(1))%coord%x + &
    patch%vertex(patch%element(elem)%vertex(2))%coord%x)/2
patch%vertex(vert1)%coord%y = &
   (patch%vertex(patch%element(elem)%vertex(1))%coord%y + &
    patch%vertex(patch%element(elem)%vertex(2))%coord%y)/2
patch%vertex(vert1)%bmark = patch%edge(patch%element(elem)%edge(3))%bmark
allocate(patch%vertex(vert1)%type(patch%system_size),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_patch_element")
   stop
endif
patch%vertex(vert1)%type = patch%edge(patch%element(elem)%edge(3))%type
patch%vertex(vert1)%grid_id = 0
allocate(patch%vertex(vert1)%grid_solution(patch%system_size,patch%num_evec), &
         patch%vertex(vert1)%solution(patch%system_size,patch%num_evec), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_patch_element")
   stop
endif
patch%vertex(vert1)%grid_solution = &
   (patch%vertex(patch%element(elem)%vertex(1))%grid_solution + &
    patch%vertex(patch%element(elem)%vertex(2))%grid_solution)/2
patch%vertex(vert1)%solution = 0.0_my_real

! New edges

paredge = patch%element(elem)%edge(3)
deg = patch%edge(paredge)%degree

! First edge

patch%edge(edge1)%vertex = (/patch%edge(paredge)%vertex(1),vert1/)
patch%edge(edge1)%degree = deg
patch%edge(edge1)%isleaf = .true.
patch%edge(edge1)%bmark = patch%edge(patch%element(elem)%edge(3))%bmark
allocate(patch%edge(edge1)%type(patch%system_size),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_patch_element")
   stop
endif
patch%edge(edge1)%type = patch%edge(patch%element(elem)%edge(3))%type
patch%edge(edge1)%grid_id = 0
if (deg > 1) then
   allocate(patch%edge(edge1)%grid_solution(deg-1,patch%system_size,patch%num_evec), &
            patch%edge(edge1)%solution(deg-1,patch%system_size,patch%num_evec),&
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%edge(edge1)%grid_solution = 0.0_my_real
   patch%edge(edge1)%solution = 0.0_my_real
else
   nullify(patch%edge(edge1)%grid_solution, &
           patch%edge(edge1)%solution)
endif

! Second edge

patch%edge(edge2)%vertex = (/patch%edge(paredge)%vertex(2),vert1/)
patch%edge(edge2)%degree = deg
patch%edge(edge2)%isleaf = .true.
patch%edge(edge2)%bmark = patch%edge(patch%element(elem)%edge(3))%bmark
allocate(patch%edge(edge2)%type(patch%system_size),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_patch_element")
   stop
endif
patch%edge(edge2)%type = patch%edge(patch%element(elem)%edge(3))%type
patch%edge(edge2)%grid_id = 0
if (deg > 1) then
   allocate(patch%edge(edge2)%grid_solution(deg-1,patch%system_size,patch%num_evec), &
            patch%edge(edge2)%solution(deg-1,patch%system_size,patch%num_evec),&
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%edge(edge2)%grid_solution = 0.0_my_real
   patch%edge(edge2)%solution = 0.0_my_real
else
   nullify(patch%edge(edge2)%grid_solution, &
           patch%edge(edge2)%solution)
endif

! Third edge

deg = patch%element(elem)%degree
patch%edge(edge3)%vertex = (/patch%element(elem)%vertex(3),vert1/)
patch%edge(edge3)%degree = deg
patch%edge(edge3)%isleaf = .true.
patch%edge(edge3)%bmark = 0
allocate(patch%edge(edge3)%type(patch%system_size),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_patch_element")
   stop
endif
patch%edge(edge3)%type = INTERIOR
patch%edge(edge3)%grid_id = 0
if (deg > 1) then
   allocate(patch%edge(edge3)%grid_solution(deg-1,patch%system_size,patch%num_evec), &
            patch%edge(edge3)%solution(deg-1,patch%system_size,patch%num_evec),&
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%edge(edge3)%grid_solution = 0.0_my_real
   patch%edge(edge3)%solution = 0.0_my_real
else
   nullify(patch%edge(edge3)%grid_solution, &
           patch%edge(edge3)%solution)
endif

! Fourth edge

if (mate /= BOUNDARY) then
   deg = patch%element(mate)%degree
   patch%edge(edge4)%vertex = (/patch%element(mate)%vertex(3),vert1/)
   patch%edge(edge4)%degree = deg
   patch%edge(edge4)%isleaf = .true.
   patch%edge(edge4)%bmark = 0
   allocate(patch%edge(edge4)%type(patch%system_size),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_patch_element")
      stop
   endif
   patch%edge(edge4)%type = INTERIOR
   patch%edge(edge4)%grid_id = 0
   if (deg > 1) then
      allocate(patch%edge(edge4)%grid_solution(deg-1,patch%system_size,patch%num_evec), &
               patch%edge(edge4)%solution(deg-1,patch%system_size,patch%num_evec), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_patch_element")
         stop
      endif
      patch%edge(edge4)%grid_solution = 0.0_my_real
      patch%edge(edge4)%solution = 0.0_my_real
   else
      nullify(patch%edge(edge4)%grid_solution, &
              patch%edge(edge4)%solution)
   endif
endif

! Grid scalars

if (mate == BOUNDARY) then
   patch%nelem = patch%nelem + 2
   patch%nedge = patch%nedge + 3
else
   patch%nelem = patch%nelem + 4
   patch%nedge = patch%nedge + 4
endif
patch%nvert = patch%nvert + 1
patch%nlev = max(patch%nlev,patch%element(child1)%level)

end subroutine h_refine_patch_element

!          --------------
subroutine h_refine_patch(patch)
!          --------------

!----------------------------------------------------
! This subroutine performs a uniform h-refinement of the entire patch.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
!----------------------------------------------------
! Local variables:

integer :: limit, elem
!----------------------------------------------------
! Begin executable code

limit = patch%nelem
do elem=1,limit
   if (patch%element(elem)%isleaf) then
      call h_refine_patch_element(patch,elem)
   endif
end do

end subroutine h_refine_patch

!          ----------------------
subroutine p_refine_patch_element(patch,elem,rule)
!          ----------------------

!----------------------------------------------------
! This subroutine p-refines element elem and enforces the edge rule.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: elem              ! The element to refine
integer, intent(in) :: rule              ! MINIMUM_RULE or MAXIMUM_RULE
!----------------------------------------------------
! Local variables:

integer :: i, newdeg, edgedeg(EDGES_PER_ELEMENT), edge, neigh
!----------------------------------------------------
! Begin executable code

! Increase the degree by one

newdeg = patch%element(elem)%degree + 1
patch%element(elem)%degree = newdeg

! For each edge, apply the edge rule

do i=1,EDGES_PER_ELEMENT
   edge = patch%element(elem)%edge(i)
   select case (rule)
   case (MINIMUM_RULE)
      neigh = patch%element(elem)%neighbor(i)
      if (neigh == BOUNDARY) then
         edgedeg(i) = newdeg
      else
         edgedeg(i) = min(newdeg,patch%element(neigh)%degree)
      endif
   case (MAXIMUM_RULE)
      neigh = patch%element(elem)%neighbor(i)
      if (neigh == BOUNDARY) then
         edgedeg(i) = newdeg
      else
         edgedeg(i) = max(newdeg,patch%element(neigh)%degree)
      endif
   case default
      ierr = PHAML_INTERNAL_ERROR
      call fatal("p_refine_patch_element: bad value for edge rule")
      stop
   end select
   patch%edge(edge)%degree = edgedeg(i)
end do

! Increase the size of solution components, if necessary

call reallocate_solution(patch,elem,newdeg,edgedeg)

end subroutine p_refine_patch_element

!          --------------
subroutine p_refine_patch(patch,rule)
!          --------------

!----------------------------------------------------
! This subroutine performs a uniform p-refinement of the entire patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: rule              ! MINIMUM_RULE or MAXIMUM_RULE
!----------------------------------------------------
! Local variables:

integer :: elem
!----------------------------------------------------
! Begin executable code

do elem=1,patch%nelem
   if (patch%element(elem)%isleaf) then
      call p_refine_patch_element(patch,elem,rule)
   endif
end do

end subroutine p_refine_patch

!          -----------
subroutine set_p_patch(patch,elem,deg,rule)
!          -----------

!----------------------------------------------------
! This routine sets the degree for element elem to be deg, and enforces the
! edge rule
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: elem              ! The element to set
integer, intent(in) :: deg               ! The degree to set elem to
integer, intent(in) :: rule              ! MINIMUM_RULE or MAXIMUM_RULE
!----------------------------------------------------
! Local variables:

integer :: iedge, edge, neigh, edgedeg(EDGES_PER_ELEMENT)
!----------------------------------------------------
! Begin executable code

! Set the element degree

patch%element(elem)%degree = deg

! Apply the edge rule

do iedge = 1,EDGES_PER_ELEMENT
   edge = patch%element(elem)%edge(iedge)
   select case (rule)
   case (MINIMUM_RULE)
      neigh = patch%element(elem)%neighbor(iedge)
      if (neigh == BOUNDARY) then
         patch%edge(edge)%degree = deg
      else
         patch%edge(edge)%degree = min(deg,patch%element(neigh)%degree)
      endif
   case (MAXIMUM_RULE)
      neigh = patch%element(elem)%neighbor(iedge)
      if (neigh == BOUNDARY) then
         patch%edge(edge)%degree = deg
      else
         patch%edge(edge)%degree = max(deg,patch%element(neigh)%degree)
      endif
   case default
      ierr = PHAML_INTERNAL_ERROR
      call fatal("set_p_patch: bad value for edge rule")
      stop
   end select
   edgedeg(iedge) = patch%edge(edge)%degree
end do

! Increase the size of solution components, if necessary

call reallocate_solution(patch,elem,deg,edgedeg)

end subroutine set_p_patch

!        ----------------------------
function get_patch_number_of_elements(patch)
!        ----------------------------

!----------------------------------------------------
! This routine returns the number of elements in the patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch ! The patch
integer :: get_patch_number_of_elements
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

get_patch_number_of_elements = patch%nelem

end function get_patch_number_of_elements

!          --------------------------
subroutine create_patch_linear_system(patch,grid,resid_prob,just_rhs, &
                                      solver_control,loc_bconds_a,loc_bconds_s)
!          --------------------------

!----------------------------------------------------
! This routine creates the linear system for a patch.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
type(grid_type), intent(in) :: grid      ! The grid
logical, intent(in) :: resid_prob        ! True if solving a residual problem
logical, intent(in) :: just_rhs          ! True if just updating right hand side
type(solver_options), intent(in), optional :: &
                          solver_control ! Passed to elemental matrix to check
                                         ! for Laplacian special code
interface
   subroutine loc_bconds_a(x,y,z,bmark,itype,c,rs,extra_rs) ! Boundary condition
   use global                                               ! subroutine with
   real(my_real), intent(in) :: x(:),y(:),z(:)              ! array of points
   integer, intent(in) :: bmark
   integer, intent(out) :: itype(:)
   real(my_real), intent(out) :: c(:,:,:),rs(:,:)
   real(my_real), intent(out), optional :: extra_rs(:,:,:)
   end subroutine loc_bconds_a
   subroutine loc_bconds_s(x,y,z,bmark,itype,c,rs)          ! Boundary condition
   use global                                               ! subroutine with
   real(my_real), intent(in) :: x,y,z                       ! a single point
   integer, intent(in) :: bmark
   integer, intent(out) :: itype(:)
   real(my_real), intent(out) :: c(:,:),rs(:)
   end subroutine loc_bconds_s
end interface
optional :: loc_bconds_a, loc_bconds_s   ! Exactly one of them should be present
!----------------------------------------------------
! Local variables:

integer :: i, j, k, deg, maxneq, astat, patch_elem, grid_elem, vert, edge, ss
integer :: degree(EDGES_PER_ELEMENT+1), bmark(EDGES_PER_ELEMENT), &
           edge_type(EDGES_PER_ELEMENT,patch%system_size)
integer :: dummy(1,1), type(patch%system_size)
real(my_real) :: rs2(patch%system_size,1), rs1(patch%system_size), &
                 c3(patch%system_size,patch%system_size,1), &
                 c2(patch%system_size,patch%system_size), &
                 extra_bcrs(patch%system_size,patch%num_evec-1,1)
real(my_real), allocatable :: lmat(:,:), lrhs(:), extra_rhs(:,:), all_rhs(:,:)
real(my_real) :: xvert(VERTICES_PER_ELEMENT), yvert(VERTICES_PER_ELEMENT)
logical :: visited_vert(patch%nvert), visited_edge(patch%nedge)
!----------------------------------------------------
! Begin executable code

! For convenience

ss = patch%system_size

! Verify exacly one of the boundary condition routines is present

if (present(loc_bconds_a) .eqv. present(loc_bconds_s)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("create_patch_linear_system: exactly one of loc_bconds_a and loc_bconds_s must be present")
   stop
endif

! Local h error indicator patch has its own routine to compute linear system

if (patch%type_of_patch == LOCAL_H_ERRIND_PATCH) then
   if (.not. present(loc_bconds_a)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("create_patch_linear_system: must have loc_bconds_a for local h error indicator")
      stop
   endif
   call special_local_h_errind_linsys(patch,grid,solver_control,loc_bconds_a)
   return
endif

! Determine the first equation associated with each patch grid entity and
! the number of equations, with and without Dirichlet rows.

call set_first_eq(patch,patch%neq,patch%n)

! Allocate memory for the linear system

allocate(patch%A(patch%neq,patch%neq),patch%b(patch%neq,patch%num_evec), &
         patch%dirichlet(patch%neq),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_patch_linear_system")
   stop
endif
patch%A = 0.0_my_real
patch%b = 0.0_my_real
patch%dirichlet = .false.

! Memory for the elemental matrix, big enough for the largest one.

deg = max(maxval(patch%element(1:patch%nelem)%degree), &
          maxval(patch%edge(1:patch%nedge)%degree))
maxneq = (3 + 3*(deg-1) + element_dof(deg))*ss
allocate(lmat(maxneq,maxneq),lrhs(maxneq),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_patch_linear_system")
   stop
endif
if (patch%num_evec > 1) then
   allocate(extra_rhs(maxneq,patch%num_evec-1),all_rhs(maxneq,patch%num_evec), &
            stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in create_patch_linear_system")
      stop
   endif
else
   allocate(all_rhs(maxneq,1),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in create_patch_linear_system")
      stop
   endif
endif

! For each patch leaf element ...

do patch_elem=1,patch%nelem
   if (.not. patch%element(patch_elem)%isleaf) cycle

! Get the vertex coordinates

   xvert = patch%vertex(patch%element(patch_elem)%vertex)%coord%x
   yvert = patch%vertex(patch%element(patch_elem)%vertex)%coord%y

! Set the degree, edge_type and bmark

   do i=1,EDGES_PER_ELEMENT
      degree(i) = patch%edge(patch%element(patch_elem)%edge(i))%degree
      edge_type(i,:) = patch%edge(patch%element(patch_elem)%edge(i))%type
      bmark(i) = patch%edge(patch%element(patch_elem)%edge(i))%bmark
   end do
   degree(EDGES_PER_ELEMENT+1) = patch%element(patch_elem)%degree

! Get the element in grid that contains this element

   grid_elem = patch%element(patch_elem)%grid_id

! Compute the elemental matrix

   if (patch%num_evec > 1) then
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, dummy, &
                            bmark, ss, 0, resid_prob, grid_elem,&
                            "p", "a", lmat, lrhs, loc_bconds_a, &
                            loc_bconds_s, solver_control=solver_control, &
                            just_rhs=just_rhs, extra_rhs=extra_rhs)
      all_rhs(:,2:) = extra_rhs
   else
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, dummy, &
                            bmark, ss, 0, resid_prob, grid_elem,&
                            "p", "a", lmat, lrhs, loc_bconds_a, &
                            loc_bconds_s, solver_control=solver_control, &
                            just_rhs=just_rhs)
   endif
   all_rhs(:,1) = lrhs

! Assemble the elemental matrix into the linear system

   if (patch%num_evec > 1) then
      call assemble(patch,patch_elem,lmat,lrhs,just_rhs,extra_rhs)
   else
      call assemble(patch,patch_elem,lmat,lrhs,just_rhs)
   endif

end do

! Set the solution and Dirichlet flags for any Dirichlet boundary vertices and
! edges

visited_vert = .false.
visited_edge = .false.

do patch_elem=1,patch%nelem
   if (.not. patch%element(patch_elem)%isleaf) cycle

! Set the vertex Dirichlet values from the boundary conditions.
! If bconds returns huge(0.0_my_real), get the solution from grid_solution.

   do i=1,VERTICES_PER_ELEMENT
      vert = patch%element(patch_elem)%vertex(i)
      if (visited_vert(vert)) cycle
      visited_vert(vert) = .true.
      if (any(patch%vertex(vert)%type == DIRICHLET)) then
         if (present(loc_bconds_a)) then
            if (patch%num_evec > 1) then
               call loc_bconds_a((/patch%vertex(vert)%coord%x/), &
                                 (/patch%vertex(vert)%coord%y/), &
                                 (/0.0_my_real/), &
                                 patch%vertex(vert)%bmark, &
                                 type, c3, rs2, extra_bcrs)
            else
               call loc_bconds_a((/patch%vertex(vert)%coord%x/), &
                                 (/patch%vertex(vert)%coord%y/), &
                                 (/0.0_my_real/), &
                                 patch%vertex(vert)%bmark, &
                                 type, c3, rs2)
            endif
         else
            call loc_bconds_s(patch%vertex(vert)%coord%x, &
                              patch%vertex(vert)%coord%y, &
                              0.0_my_real, &
                              patch%vertex(vert)%bmark, &
                              type, c2, rs1)
            do j=1,patch%num_evec-1
               extra_bcrs(:,j,1) = rs1(:)
            end do
         endif
         do j=1,ss
            if (patch%vertex(vert)%type(j) == DIRICHLET) then
               if (present(loc_bconds_a)) then
                  if (rs2(j,1) == huge(0.0_my_real)) then
                     patch%vertex(vert)%solution(j,1) = &
                        patch%vertex(vert)%grid_solution(j,1)
                  else
                     patch%vertex(vert)%solution(j,1) = rs2(j,1)
                  endif
               else
                  if (rs1(j) == huge(0.0_my_real)) then
                     patch%vertex(vert)%solution(j,1) = &
                        patch%vertex(vert)%grid_solution(j,1)
                  else
                     patch%vertex(vert)%solution(j,1) = rs1(j)
                  endif
               endif
               if (patch%num_evec > 1) then
                  patch%vertex(vert)%solution(j,2:) = extra_bcrs(j,:,1)
               endif
               if (resid_prob .and. patch%vertex(vert)%bmark /= huge(0)) then
                  patch%vertex(vert)%solution(j,:) = &
                          patch%vertex(vert)%solution(j,:) - &
                          patch%vertex(vert)%grid_solution(j,:)
               endif
               patch%dirichlet(patch%vertex(vert)%first_eq+j-1) = .true.
               patch%b(patch%vertex(vert)%first_eq+j-1,:) = &
                  patch%vertex(vert)%solution(j,:)
            endif
         end do
      endif
   end do

! Set the edge Dirichlet values

   do i=1,EDGES_PER_ELEMENT
      edge = patch%element(patch_elem)%edge(i)
      if (patch%edge(edge)%first_eq == 0) cycle
      if (visited_edge(edge)) cycle
      visited_edge(edge) = .true.
      if (any(patch%edge(edge)%type == DIRICHLET)) then
         do j=1,ss
            if (patch%edge(edge)%type(j) == DIRICHLET) then
               call patch_edge_exact(patch,edge,j)
               do k=1,patch%edge(edge)%degree-1
                  patch%dirichlet(patch%edge(edge)%first_eq+j-1+ss*(k-1))=.true.
                  patch%b(patch%edge(edge)%first_eq+j-1+ss*(k-1),:) = &
                     patch%edge(edge)%solution(k,j,:)
               end do
            endif
         end do
      endif
   end do

end do

deallocate(lmat,lrhs,all_rhs)
if (patch%num_evec > 1) deallocate(extra_rhs)

! Set the vectors of indices of Dirichlet and non-Dirichlet rows

allocate(patch%nondir_rows(patch%n),patch%dir_rows(patch%neq-patch%n), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_patch_linear_system")
   stop
endif
j = 0
k = 0
do i=1,patch%neq
   if (patch%dirichlet(i)) then
      k = k + 1
      patch%dir_rows(k) = i
   else
      j = j + 1
      patch%nondir_rows(j) = i
   endif
end do
if (j /= patch%n) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("create_patch_linear_system: mismatch in number of Dirichlet rows")
   stop
endif

end subroutine create_patch_linear_system

!          -----------------------------
subroutine special_local_h_errind_linsys(patch,grid,solver_control,loc_bconds_a)
!          -----------------------------

!----------------------------------------------------
! This routine computes the linear system for a patch for the local h
! error indicator.  The patch should consist of the primary element and mate
! and no halo, and be h-refined once.  The linear system solves the residual
! problem using the red bases of the nodal basis.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
type(grid_type), intent(in) :: grid      ! The grid
type(solver_options), intent(in), optional :: &
                          solver_control ! Passed to elemental matrix to check
                                         ! for Laplacian special code
interface
   subroutine loc_bconds_a(x,y,z,bmark,itype,c,rs,extra_rs) ! Boundary condition
   use global                                               ! subroutine with
   real(my_real), intent(in) :: x(:),y(:),z(:)              ! array of points
   integer, intent(in) :: bmark
   integer, intent(out) :: itype(:)
   real(my_real), intent(out) :: c(:,:,:),rs(:,:)
   real(my_real), intent(out), optional :: extra_rs(:,:,:)
   end subroutine loc_bconds_a
end interface
!----------------------------------------------------
! Local variables:

real(my_real) :: xvert(3), yvert(3), alpha, x, y, solut(patch%system_size,1,1),&
                 c(patch%system_size,patch%system_size), rs(patch%system_size)
real(my_real), allocatable :: elem_mat(:,:),elem_rs(:,:)
integer :: elem, lolim, deg, degree(EDGES_PER_ELEMENT+1), nred_elem, nred_full,&
           elem_n, astat, ss, i, i2, j, k, k1, k2, k3, k4, &
           bmark(EDGES_PER_ELEMENT), itype(grid%system_size), nev, &
           edge_type(EDGES_PER_ELEMENT,patch%system_size), face_type(1,1)
integer, allocatable :: renum(:)
!----------------------------------------------------
! Begin executable code

! Make sure the patch was created correctly.  The number of elements is
! 3 if there is no mate (i.e. refinement edge is on the boundary)
!   1 - primary
!   2,3 - children of primary
! 6 if the mate is compatibly divisible
!   1 - primary
!   2 - mate
!   3,4 - children of primary
!   5,6 - children of mate
! 8 if the mate is not compatibly divisible
!   1 - primary
!   2 - parent of mate
!   3,4 - mate and sibling of mate
!   5,6 - children of primary
!   7,8 - children of mate

if (patch%nelem /= 3 .and. patch%nelem /= 6 .and. patch%nelem /= 8) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("special_local_h_errind_linsys: number of patch elements is not 3, 6 or 8")
   stop
endif

! The last vertex was created when the primary was refined.  Make sure the
! children of the primary, and children of the mate if it exists, have the
! last vertex as their 3rd vertex.

if (patch%nelem == 3) then
   lolim = 2
else
   lolim = patch%nelem-3
endif
do elem=lolim,patch%nelem
   if (patch%element(elem)%vertex(3) /= patch%nvert) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("special_local_h_errind_linsys: 3rd vertex of child is not the last vertex")
      stop
   endif
end do

! First_eq's and neq's are not used in this special case

patch%vertex%first_eq = 0
patch%vertex%neq = 0
patch%edge%first_eq = 0
patch%edge%neq = 0
patch%element%first_eq = 0
patch%element%neq = 0

! Convenience variables

ss = grid%system_size
nev = max(1,grid%num_eval)

! Degree of the elements

deg = patch%element(1)%degree
degree = deg

! Number of red nodal bases over the refinement of triangles of this degree,
! both individual elements and all children

if (2*(deg/2) == deg) then
   nred_elem = (deg*(deg+2))/4
else
   nred_elem = ((deg+1)*(deg+1))/4
endif
if (patch%nelem == 3) then
   nred_full = (deg*(deg+1))/2
else
   nred_full = deg**2
endif

! Number of equations

elem_n = nred_elem*ss
patch%neq = nred_full*ss
patch%n = patch%neq

! Allocate and initialize matrices, etc.

allocate(elem_mat(elem_n,elem_n), elem_rs(elem_n,nev), renum(elem_n), &
         patch%A(patch%neq,patch%neq), patch%b(patch%neq,nev), &
         patch%dirichlet(patch%neq), stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in special_local_h_errind_linsys")
   stop
endif

elem_mat = 0.0_my_real
elem_rs = 0.0_my_real
patch%A = 0.0_my_real
patch%b = 0.0_my_real
patch%dirichlet = .false.

! For each of the children of the primary and mate, compute the elemental matrix
! and right side, and assemble

! first child

if (patch%nelem == 3) then
   elem = 2
else
   elem = patch%nelem - 3
endif
xvert = patch%vertex(patch%element(elem)%vertex)%coord%x
yvert = patch%vertex(patch%element(elem)%vertex)%coord%y

edge_type(1,:) = INTERIOR
bmark(1) = 0
if (patch%nelem == 3) then
   edge_type(2,:) = patch%edge(patch%element(elem)%edge(2))%type
   bmark(2) = patch%edge(patch%element(elem)%edge(2))%bmark
else
   edge_type(2,:) = INTERIOR
   bmark(2) = 0
endif
edge_type(3,:) = DIRICHLET
bmark(3) = huge(0)

if (nev > 1) then
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark, ss, 0, .true., patch%element(1)%grid_id, "n", &
                         "r", elem_mat, elem_rs(:,1), &
                         loc_bconds_a=loc_bconds_a, &
                         solver_control=solver_control, extra_rhs=elem_rs(:,2:))
else
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark, ss, 0, .true., patch%element(1)%grid_id, "n", &
                         "r", elem_mat, elem_rs(:,1), &
                         solver_control=solver_control, &
                         loc_bconds_a=loc_bconds_a)
endif

patch%A(1:elem_n,1:elem_n) = elem_mat
patch%b(1:elem_n,:) = elem_rs

! second child

if (patch%nelem == 3) then
   elem = 3
else
   elem = patch%nelem - 2
endif
xvert = patch%vertex(patch%element(elem)%vertex)%coord%x
yvert = patch%vertex(patch%element(elem)%vertex)%coord%y

edge_type(1,:) = INTERIOR
bmark(1) = 0
if (patch%nelem == 3) then
   edge_type(2,:) = patch%edge(patch%element(elem)%edge(2))%type
   bmark(2) = patch%edge(patch%element(elem)%edge(2))%bmark
else
   edge_type(2,:) = INTERIOR
   bmark(2) = 0
endif
edge_type(3,:) = DIRICHLET
bmark(3) = huge(0)

if (nev > 1) then
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark, ss, 0, .true., patch%element(1)%grid_id, "n", &
                         "r", elem_mat, elem_rs(:,1), &
                         loc_bconds_a=loc_bconds_a, &
                         solver_control=solver_control, extra_rhs=elem_rs(:,2:))
else
   call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                         bmark, ss, 0, .true., patch%element(1)%grid_id, "n",  &
                         "r", elem_mat, elem_rs(:,1), &
                         solver_control=solver_control, &
                         loc_bconds_a=loc_bconds_a)
endif

k1 = 1
k2 = elem_n/ss + 1
k3 = deg
do i=1,(deg+1)/2
   do j=1,deg-2*i+1
      renum(k1:k1+ss-1) = (/ (ss*(k2-1)+i2, i2=1,ss) /)
      k1 = k1+ss
      k2 = k2+1
   end do
   renum(k1:k1+ss-1) = (/ (ss*(k3-1)+i2, i2=1,ss) /)
   k1 = k1+ss
   k3 = k3+deg-2*i
end do

do i=1,elem_n
   do j=1,elem_n
      patch%A(renum(i),renum(j)) = patch%A(renum(i),renum(j)) + elem_mat(i,j)
   end do
   patch%b(renum(i),:) = patch%b(renum(i),:) + elem_rs(i,:)
end do

if (patch%nelem /= 3) then

! third child

   elem = patch%nelem - 1
   xvert = patch%vertex(patch%element(elem)%vertex)%coord%x
   yvert = patch%vertex(patch%element(elem)%vertex)%coord%y

   edge_type(1,:) = INTERIOR
   bmark(1) = 0
   edge_type(2,:) = INTERIOR
   bmark(2) = 0
   edge_type(3,:) = DIRICHLET
   bmark(3) = huge(0)

   if (nev > 1) then
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, 0, .true., patch%element(2)%grid_id, &
                            "n", "r", elem_mat, elem_rs(:,1), &
                            loc_bconds_a=loc_bconds_a, &
                            solver_control=solver_control, &
                            extra_rhs=elem_rs(:,2:))
   else
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, 0, .true., patch%element(2)%grid_id, &
                            "n", "r", elem_mat, elem_rs(:,1), &
                            solver_control=solver_control, &
                            loc_bconds_a=loc_bconds_a)
   endif

   k1 = 1
   k2 = (deg*(deg+1))/2 + 1
   k3 = 1
   do i=1,(deg+1)/2
      renum(k1:k1+ss-1) = (/ (ss*(k3-1)+i2, i2=1,ss) /)
      k1 = k1+ss
      k3 = k3+deg-2*i+2
      do j=1,deg-2*i+1
         renum(k1:k1+ss-1) = (/ (ss*(k2-1)+i2, i2=1,ss) /)
         k1 = k1+ss
         k2 = k2+1
      end do
   end do

   do i=1,elem_n
      do j=1,elem_n
        patch%A(renum(i),renum(j)) = patch%A(renum(i),renum(j))+elem_mat(i,j)
      end do
      patch%b(renum(i),:) = patch%b(renum(i),:) + elem_rs(i,:)
   end do

! fourth child

   elem = patch%nelem
   xvert = patch%vertex(patch%element(elem)%vertex)%coord%x
   yvert = patch%vertex(patch%element(elem)%vertex)%coord%y

   edge_type(1,:) = INTERIOR
   bmark(1) = 0
   edge_type(2,:) = INTERIOR
   bmark(2) = 0
   edge_type(3,:) = DIRICHLET
   bmark(3) = huge(0)

   if (nev > 1) then
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, 0, .true., patch%element(2)%grid_id, &
                            "n", "r", elem_mat, elem_rs(:,1), &
                            loc_bconds_a=loc_bconds_a, &
                            solver_control=solver_control, &
                            extra_rhs=elem_rs(:,2:))
   else
      call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, 0, .true., patch%element(2)%grid_id, &
                            "n", "r", elem_mat, elem_rs(:,1), &
                            solver_control=solver_control, &
                            loc_bconds_a=loc_bconds_a)
   endif

   k1 = 1
   if (2*(deg/2) == deg) then
      k2 = (deg*(3*deg+2))/4 + 1
   else
      k2 = ((deg+1)*(3*deg-1))/4 + 1
   endif
   k3 = elem_n/ss + 1
   k4 = (deg*(deg+1))/2 + deg-1
   do i=1,deg/2
      renum(k1:k1+ss-1) = (/ (ss*(k3-1)+i2, i2=1,ss) /)
      k1 = k1+ss
      k3 = k3+deg-2*i+1
      do j=1,deg-2*i
         renum(k1:k1+ss-1) = (/ (ss*(k2-1)+i2, i2=1,ss) /)
         k1 = k1+ss
         k2 = k2+1
      end do
      renum(k1:k1+ss-1) = (/ (ss*(k4-1)+i2, i2=1,ss) /)
      k1 = k1+ss
      k4 = k4+deg-2*i-1
   end do
   if (2*(deg/2) /= deg) then
      renum(k1:k1+ss-1) = (/ (elem_n-ss+i2, i2=1,ss) /)
   endif

   do i=1,elem_n
      do j=1,elem_n
        patch%A(renum(i),renum(j)) = patch%A(renum(i),renum(j))+elem_mat(i,j)
      end do
      patch%b(renum(i),:) = patch%b(renum(i),:) + elem_rs(i,:)
   end do

else ! mate is BOUNDARY

! Replace rows corresponding to Dirichlet nodes with an identity row and
! the right side with the error

! Points in first triangle

   k1 = 1
   do i=1,(deg+1)/2
      alpha = real(2*i-1,my_real)/real(2*deg,my_real)
      x = (1-alpha)*patch%vertex(patch%element(1)%vertex(1))%coord%x + &
             alpha *patch%vertex(patch%element(1)%vertex(2))%coord%x
      y = (1-alpha)*patch%vertex(patch%element(1)%vertex(1))%coord%y + &
             alpha *patch%vertex(patch%element(1)%vertex(2))%coord%y
      if (any(patch%edge(patch%element(1)%edge(3))%type == DIRICHLET)) then
         call evaluate_soln_local(grid,(/x/),(/y/),patch%element(1)%grid_id, &
                                  (/(j,j=1,ss)/),(/1/),solut)
         call my_bconds(x,y,0.0_my_real, &
                        patch%edge(patch%element(1)%edge(3))%bmark,itype,c,rs)
      endif
      do j=1,ss
         if (patch%edge(patch%element(1)%edge(3))%type(j) == DIRICHLET) then
            patch%b(k1+j-1,:) = rs(j) - solut(j,1,1)
            patch%A(k1+j-1,:) = 0.0_my_real
            patch%A(k1+j-1,k1+j-1) = 1.0_my_real
            patch%dirichlet(k1+j-1) = .true.
            patch%n = patch%n - 1
         endif
      end do
      k1 = k1 + ss*(deg-2*(i-1))
   end do

! Points in second triangle

   k1 = elem_n + 1
   do i=1,deg/2
      alpha = real(2*i-1,my_real)/real(2*deg,my_real)
      x = (1-alpha)*patch%vertex(patch%element(1)%vertex(2))%coord%x + &
             alpha *patch%vertex(patch%element(1)%vertex(1))%coord%x
      y = (1-alpha)*patch%vertex(patch%element(1)%vertex(2))%coord%y + &
             alpha *patch%vertex(patch%element(1)%vertex(1))%coord%y
      if (any(patch%edge(patch%element(1)%edge(3))%type == DIRICHLET)) then
         call evaluate_soln_local(grid,(/x/),(/y/),patch%element(1)%grid_id, &
                                  (/(j,j=1,ss)/),(/1/),solut)
         call my_bconds(x,y,0.0_my_real, &
                        patch%edge(patch%element(1)%edge(3))%bmark,itype,c,rs)
      endif
      do j=1,ss
         if (patch%edge(patch%element(elem)%edge(3))%type(j) == DIRICHLET) then
            patch%b(k1+j-1,:) = rs(j) - solut(j,1,1)
            patch%A(k1+j-1,:) = 0.0_my_real
            patch%A(k1+j-1,k1+j-1) = 1.0_my_real
            patch%dirichlet(k1+j-1) = .true.
            patch%n = patch%n - 1
         endif
      end do
      k1 = k1 + ss*(deg+1-2*i)
   end do

endif

! Set the vectors of indices of Dirichlet and non-Dirichlet rows

allocate(patch%nondir_rows(patch%n),patch%dir_rows(patch%neq-patch%n), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in create_patch_linear_system")
   stop
endif
j = 0
k = 0
do i=1,patch%neq
   if (patch%dirichlet(i)) then
      k = k + 1
      patch%dir_rows(k) = i
   else
      j = j + 1
      patch%nondir_rows(j) = i
   endif
end do
if (j /= patch%n) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("special_local_h_errind_linsys: mismatch in number of Dirichlet rows")
   stop
endif

! clean up memory

deallocate(elem_mat,elem_rs,renum)

end subroutine special_local_h_errind_linsys

!          ----------------------------
subroutine destroy_patch_linear_system(patch)
!          ----------------------------

!----------------------------------------------------
! This routine destroys the linear system for a patch.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

! Deallocate patch linear system memory

if (associated(patch%A)) deallocate(patch%A)
if (associated(patch%LU)) deallocate(patch%LU)
if (associated(patch%b)) deallocate(patch%b)
if (associated(patch%x)) deallocate(patch%x)
if (associated(patch%ipiv)) deallocate(patch%ipiv)
if (associated(patch%dirichlet)) deallocate(patch%dirichlet)
if (associated(patch%nondir_rows)) deallocate(patch%nondir_rows)
if (associated(patch%dir_rows)) deallocate(patch%dir_rows)

! There is no linear system any more

patch%neq = 0
patch%n   = 0

end subroutine destroy_patch_linear_system

!          -------------------------
subroutine solve_patch_linear_system(patch)
!          -------------------------

!----------------------------------------------------
! This routine solves the linear system for a patch.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
!----------------------------------------------------
! Local variables:

integer :: info, astat
!----------------------------------------------------
! Begin executable code

! If there are no equations, there is nothing to do

if (patch%n == 0) return

! Allocate solution vector

if (.not. associated(patch%x)) then
   allocate(patch%x(patch%n,patch%num_evec),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in solve_patch_linear_system")
      stop
   endif
endif

! If there is one equation, just set the solution

if (patch%n == 1) then
   patch%x(1,:) = patch%b(patch%nondir_rows(1),:) / &
                  patch%A(patch%nondir_rows(1),patch%nondir_rows(1))
   call linsys_solution_to_grid(patch)
   return
endif

! Use LAPACK routines for a full dense matrix

! If the matrix has not already been factored, allocate the memory for the
! factored matrix, copy the matrix to it while removing Dirichlet rows, and
! factor it

if (.not. associated(patch%LU)) then

   allocate(patch%LU(patch%n,patch%n),patch%ipiv(patch%n),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in solve_patch_linear_system")
      stop
   endif

   patch%LU = patch%A(patch%nondir_rows,patch%nondir_rows)

   if (my_real == kind(1.0)) then
      call sgetrf(patch%n,patch%n,patch%LU,patch%n,patch%ipiv,info)
   elseif (my_real == kind(1.0d0)) then
      call dgetrf(patch%n,patch%n,patch%LU,patch%n,patch%ipiv,info)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("lapack requires my_real is either default single or double precision")
   endif
   if (info /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("lapack factorization failed in solve_patch_linear_system")
      stop
   endif

endif

! Copy the right hand side to the solution, removing Dirichlet boundary
! conditions

patch%x = patch%b(patch%nondir_rows,:) - &
   matmul(patch%A(patch%nondir_rows,patch%dir_rows),patch%b(patch%dir_rows,:))

! Solve the linear system

if (my_real == kind(1.0)) then
   call sgetrs("N",patch%n,patch%num_evec,patch%LU,patch%n,patch%ipiv,patch%x, &
               patch%n,info)
else
   call dgetrs("N",patch%n,patch%num_evec,patch%LU,patch%n,patch%ipiv,patch%x, &
               patch%n,info)
endif
if (info /= 0) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("lapack solution failed in solve_patch_linear_system")
   stop
endif

! Copy the patch linear system solution to the patch grid solution

call linsys_solution_to_grid(patch)

end subroutine solve_patch_linear_system

!        ---------------------------
function get_patch_linear_system_dof(patch,nodirich)
!        ---------------------------

!----------------------------------------------------
! This routine returns the number of degrees of freedom in the linear system.
! If nodirich is present and true, Dirichlet degrees are not included.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch
logical, optional, intent(in) :: nodirich
integer :: get_patch_linear_system_dof

!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (present(nodirich)) then
   if (nodirich) then
      get_patch_linear_system_dof = patch%n
   else
      get_patch_linear_system_dof = patch%neq
   endif
else
   get_patch_linear_system_dof = patch%neq
endif

end function get_patch_linear_system_dof

!          ---------------
subroutine norm_patch_diff(patch,energy,Linf,L2,nodirich)
!          ---------------

!----------------------------------------------------
! This routine returns norms of the difference between the grid solution
! and the patch solution.
! If nodirich is present and true, the Dirichlet entries are omitted.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch
real(my_real), intent(out), optional :: energy(:), &      ! dimension (num_evec)
                            Linf(:,:),L2(:,:) ! dimension (system_size,num_evec)
logical, intent(in), optional :: nodirich
!----------------------------------------------------
! Local variables:

integer :: ss, nev, i, j, k, l, ivert, vert, row
real(my_real) :: diff(patch%neq,patch%num_evec), Adiff(patch%n,patch%num_evec),&
                 midpoint_val(patch%system_size,patch%num_evec), &
                 xvert(VERTICES_PER_ELEMENT), yvert(VERTICES_PER_ELEMENT)
real(my_real) :: area
logical :: yesdirich
!----------------------------------------------------
! Begin executable code

! Cannot compute the difference between the solution and the grid solution
! for a special local h error indicator patch, because it uses the nodal
! basis.

if (patch%type_of_patch == LOCAL_H_ERRIND_PATCH) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("norm_patch_diff: cannot compute difference with local h error indicator patch")
   stop
endif

! Convenience variables

ss = patch%system_size
nev = patch%num_evec
if (present(nodirich)) then
   yesdirich = .not. nodirich
else
   yesdirich = .true.
endif

! Create the vector of differences between solution and grid_solution,
! approximate the L-infinity norm as the maximum at the vertices, and
! approximate the L2 norm using midpoint quadrature with the average of the
! function at the vertices.

if (present(Linf)) Linf = 0
if (present(L2)) L2 = 0

do i=1,patch%nvert
   diff(patch%vertex(i)%first_eq: &
        patch%vertex(i)%first_eq+patch%vertex(i)%neq-1,:) = &
        patch%vertex(i)%solution - patch%vertex(i)%grid_solution
   if (present(Linf)) then
      do j=1,patch%vertex(i)%neq
         if (yesdirich .or. &
             .not. patch%dirichlet(patch%vertex(i)%first_eq+j-1)) then
            Linf(j,:) = max(abs(diff(patch%vertex(i)%first_eq+j-1,:)),Linf(j,:))
         endif
      end do
   endif
end do

do i=1,patch%nedge
   if (patch%edge(i)%first_eq == 0) cycle
   do j=1,patch%edge(i)%neq/ss
      k = patch%edge(i)%first_eq + (j-1)*ss
      diff(k:k+ss-1,:) = patch%edge(i)%solution(j,:,:) - &
                         patch%edge(i)%grid_solution(j,:,:)
   end do
end do

do i=1,patch%nelem
   if (.not. patch%element(i)%isleaf) cycle
   if (patch%element(i)%first_eq /= 0) then
      do j=1,patch%element(i)%neq/ss
         k = patch%element(i)%first_eq + (j-1)*ss
         diff(k:k+ss-1,:) = patch%element(i)%solution(j,:,:) - &
                            patch%element(i)%grid_solution(j,:,:)
      end do
   endif
   if (present(L2)) then
      midpoint_val = 0
      do ivert=1,VERTICES_PER_ELEMENT
         vert = patch%element(i)%vertex(ivert)
         xvert(ivert) = patch%vertex(vert)%coord%x
         yvert(ivert) = patch%vertex(vert)%coord%y
         do row=1,patch%vertex(vert)%neq
            if (yesdirich .or. &
                .not.patch%dirichlet(patch%vertex(vert)%first_eq+row-1)) then
               midpoint_val(row,:) = midpoint_val(row,:) + &
                                     diff(patch%vertex(vert)%first_eq+row-1,:)
            endif
         end do
      end do
      midpoint_val = (midpoint_val/VERTICES_PER_ELEMENT)**2
      area = abs(xvert(1)*(yvert(2)-yvert(3)) + &
                 xvert(2)*(yvert(3)-yvert(1)) + &
                 xvert(3)*(yvert(1)-yvert(2))) / 2
      L2 = L2 + midpoint_val*area
   endif
end do

if (present(L2)) L2 = sqrt(L2)

! Compute sqrt(diff^T A diff)

if (present(energy)) then
   if (yesdirich) then
      Adiff = matmul(patch%A,diff)
      do i=1,patch%num_evec
         energy(i)=sqrt(dot_product(diff(:,i),Adiff(:,i)))
      end do
   else
      Adiff(1:patch%n,:) = matmul(patch%A(patch%nondir_rows,patch%nondir_rows),&
                                  diff(patch%nondir_rows,:))
      do i=1,patch%num_evec
         energy(i)=sqrt(dot_product(diff(patch%nondir_rows,i), &
                                    Adiff(1:patch%n,i)))
      end do
   endif
endif

end subroutine norm_patch_diff

!          ---------------
subroutine norm_patch_soln(patch,energy,Linf,L2,nodirich)
!          ---------------

!----------------------------------------------------
! This routine returns norms of the solution.
! If nodirich is present and true, the Dirichlet entries are omitted.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch
real(my_real), intent(out), optional :: energy(:), &      ! dimension (num_evec)
                            Linf(:,:),L2(:,:) ! dimension (system_size,num_evec)
logical, intent(in), optional :: nodirich
!----------------------------------------------------
! Local variables:

integer :: ss, nev, i, j, k, l, ivert, vert, row
real(my_real) :: soln(patch%neq,patch%num_evec), &
                 Asoln(patch%neq,patch%num_evec), &
                 midpoint_val(patch%system_size,patch%num_evec), &
                 xvert(VERTICES_PER_ELEMENT), yvert(VERTICES_PER_ELEMENT)
real(my_real) :: area
logical :: yesdirich
!----------------------------------------------------
! Begin executable code

! Convenience variables

ss = patch%system_size
nev = patch%num_evec
if (present(nodirich)) then
   yesdirich = .not. nodirich
else
   yesdirich = .true.
endif

if (present(Linf)) Linf = 0
if (present(L2)) L2 = 0

select case(patch%type_of_patch)

case (DEFAULT_PATCH_TYPE)

! Create the vector containing the solution, approximate the L-infinity norm as
! the maximum at the vertices, and approximate the L2 norm using midpoint
! quadrature with the average of the function at the vertices.

   do i=1,patch%nvert
      soln(patch%vertex(i)%first_eq: &
           patch%vertex(i)%first_eq+patch%vertex(i)%neq-1,:) = &
           patch%vertex(i)%solution
      if (present(Linf)) then
         do j=1,patch%vertex(i)%neq
            if (yesdirich .or. &
                .not. patch%dirichlet(patch%vertex(i)%first_eq+j-1)) then
               Linf(j,:) = &
                  max(abs(soln(patch%vertex(i)%first_eq+j-1,:)),Linf(j,:))
            endif
         end do
      endif
   end do

   do i=1,patch%nedge
      if (patch%edge(i)%first_eq == 0) cycle
      do j=1,patch%edge(i)%neq/ss
         k = patch%edge(i)%first_eq + (j-1)*ss
         soln(k:k+ss-1,:) = patch%edge(i)%solution(j,:,:)
      end do
   end do

   do i=1,patch%nelem
      if (.not. patch%element(i)%isleaf) cycle
      if (patch%element(i)%first_eq /= 0) then
         do j=1,patch%element(i)%neq/ss
            k = patch%element(i)%first_eq + (j-1)*ss
            soln(k:k+ss-1,:) = patch%element(i)%solution(j,:,:)
         end do
      endif
      if (present(L2)) then
         midpoint_val = 0
         do ivert=1,VERTICES_PER_ELEMENT
            vert = patch%element(i)%vertex(ivert)
            xvert(ivert) = patch%vertex(vert)%coord%x
            yvert(ivert) = patch%vertex(vert)%coord%y
            do row=1,patch%vertex(vert)%neq
               if (yesdirich .or. &
                   .not.patch%dirichlet(patch%vertex(vert)%first_eq+row-1)) then
                  midpoint_val(row,:) = midpoint_val(row,:) + &
                                       soln(patch%vertex(vert)%first_eq+row-1,:)
               endif
            end do
         end do
         midpoint_val = (midpoint_val/VERTICES_PER_ELEMENT)**2
         area = abs(xvert(1)*(yvert(2)-yvert(3)) + &
                    xvert(2)*(yvert(3)-yvert(1)) + &
                    xvert(3)*(yvert(1)-yvert(2))) / 2
         L2 = L2 + midpoint_val*area
      endif
   end do

case (LOCAL_H_ERRIND_PATCH)

! For the local h error indicator, copy the solution from x and insert 0 for
! the Dirichlet boundary nodes.  Take the maximum of the solution vector as
! the L infinity norm.  For the L2 norm, use the maximum as the function
! value in the L2 integral over the primary element (patch element 1).

   if (.not. associated(patch%b)) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("norm_patch_soln: right side vector b has not been allocated")
      stop
   endif
   soln = 0
   if (associated(patch%x)) then
      soln(patch%nondir_rows,:) = patch%x
   endif
   soln(patch%dir_rows,:) = patch%b(patch%dir_rows,:)
   if (present(Linf)) then
      do k=1,nev
         do j=1,ss
            Linf(j,k) = maxval(abs(soln(j:patch%neq:ss,k)))
         end do
      end do
   endif
   if (present(L2)) then
      xvert = patch%vertex(patch%element(1)%vertex)%coord%x
      yvert = patch%vertex(patch%element(1)%vertex)%coord%y
      area = abs(xvert(1)*(yvert(2)-yvert(3)) + &
                 xvert(2)*(yvert(3)-yvert(1)) + &
                 xvert(3)*(yvert(1)-yvert(2))) / 2
      do k=1,nev
         do j=1,ss
            L2(j,k) = area*maxval(abs(soln(j:patch%neq:ss,k)))**2
         end do
      end do
   endif

end select

if (present(L2)) L2 = sqrt(L2)

! Compute sqrt(soln^T A soln)

if (present(energy)) then
   if (yesdirich) then
      Asoln = matmul(patch%A,soln)
      do i=1,patch%num_evec
         energy(i)=sqrt(dot_product(soln(:,i),Asoln(:,i)))
      end do
   else
      Asoln(1:patch%n,:) = matmul(patch%A(patch%nondir_rows,patch%nondir_rows),&
                                  soln(patch%nondir_rows,:))
      do i=1,patch%num_evec
         energy(i)=sqrt(dot_product(soln(patch%nondir_rows,i), &
                                    Asoln(1:patch%n,i)))
      end do
   endif
endif

end subroutine norm_patch_soln

!          ----------------
subroutine print_patch_grid(patch)
!          ----------------

!----------------------------------------------------
! This routine prints the grid components of a patch.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch
!----------------------------------------------------
! Local variables:

integer :: i
!----------------------------------------------------
! Begin executable code

write(outunit,*)
write(outunit,*) "-------------- patch ---------------"
write(outunit,*)
write(outunit,*) "Number of levels ",patch%nlev
write(outunit,*)
write(outunit,*) "Number of elements ",patch%nelem
write(outunit,*)
do i=1,patch%nelem
   write(outunit,*) "Element ",i," is grid element ",patch%element(i)%grid_id
   write(outunit,*)
   write(outunit,*) "Vertices ",patch%element(i)%vertex
   write(outunit,*) "Edges    ",patch%element(i)%edge
   write(outunit,*) "Neighbors",patch%element(i)%neighbor
   write(outunit,*) "degree   ",patch%element(i)%degree
   write(outunit,*) "level    ",patch%element(i)%level
   write(outunit,*) "isleaf   ",patch%element(i)%isleaf
   if (associated(patch%element(i)%solution)) then
      write(outunit,*) "solution ",patch%element(i)%solution
   else
      write(outunit,*) "solution not allocated"
   endif
   if (associated(patch%element(i)%grid_solution)) then
      write(outunit,*) "grid_solution ",patch%element(i)%grid_solution
   else
      write(outunit,*) "grid_solution not allocated"
   endif
   write(outunit,*)
end do
write(outunit,*) "Number of edges ",patch%nedge
write(outunit,*)
do i=1,patch%nedge
   write(outunit,*) "Edge ",i," is grid edge ",patch%edge(i)%grid_id
   write(outunit,*)
   write(outunit,*) "Vertices ",patch%edge(i)%vertex
   write(outunit,*) "degree   ",patch%edge(i)%degree
   write(outunit,*) "isleaf   ",patch%edge(i)%isleaf
   write(outunit,*) "bmark    ",patch%edge(i)%bmark
   write(outunit,*) "type     ",patch%edge(i)%type
   if (associated(patch%edge(i)%solution)) then
      write(outunit,*) "solution ",patch%edge(i)%solution
   else
      write(outunit,*) "solution not associated"
   endif
   if (associated(patch%edge(i)%grid_solution)) then
      write(outunit,*) "grid_solution ",patch%edge(i)%grid_solution
   else
      write(outunit,*) "grid_solution not associated"
   endif
   write(outunit,*)
end do
write(outunit,*) "Number of vertices ",patch%nvert
write(outunit,*)
do i=1,patch%nvert
   write(outunit,*) "Vertex ",i," is grid vertex ",patch%vertex(i)%grid_id
   write(outunit,*)
   write(outunit,*) "Coordinate ",patch%vertex(i)%coord
   write(outunit,*) "bmark    ",patch%vertex(i)%bmark
   write(outunit,*) "type     ",patch%vertex(i)%type
   if (associated(patch%vertex(i)%solution)) then
      write(outunit,*) "solution ",patch%vertex(i)%solution
   else
      write(outunit,*) "solution not associated"
   endif
   if (associated(patch%vertex(i)%grid_solution)) then
      write(outunit,*) "grid_solution ",patch%vertex(i)%grid_solution
   else
      write(outunit,*) "grid_solution not associated"
   endif
   write(outunit,*)
end do
write(outunit,*) "------------ end patch -------------"
write(outunit,*)
end subroutine print_patch_grid

!          -------------------------
subroutine print_patch_linear_system(patch)
!          -------------------------

!----------------------------------------------------
! This routine prints the linear system of patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(in) :: patch
!----------------------------------------------------
! Local variables:

integer :: i
!----------------------------------------------------
! Begin executable code

write(outunit,*)
write(outunit,*) "-------------- patch linear system ---------------"
write(outunit,*)
write(outunit,*) "Number of equations ",patch%neq
write(outunit,*)
if (associated(patch%A)) then
   write(outunit,*) "Matrix A "
   write(outunit,*)
   do i=1,patch%neq
      write(outunit,*) patch%A(:,i)
      write(outunit,*)
   end do
endif
if (associated(patch%b) .and. associated(patch%dirichlet)) then
   write(outunit,*) "Right hand side b and Dirichlet flag"
   write(outunit,*)
   do i=1,patch%neq
      write(outunit,*) patch%b(i,:),patch%dirichlet(i)
   end do
   write(outunit,*)
endif
if (associated(patch%dir_rows)) then
   write(outunit,*) "List of Dirichlet rows"
   write(outunit,*) patch%dir_rows
endif
if (associated(patch%nondir_rows)) then
   write(outunit,*) "List of non-Dirichlet rows"
   write(outunit,*) patch%nondir_rows
endif
if (associated(patch%LU)) then
   write(outunit,*) "Factored matrix LU"
   write(outunit,*)
   do i=1,patch%n
      write(outunit,*) patch%LU(:,i)
      write(outunit,*)
   end do
endif
if (associated(patch%ipiv)) then
   write(outunit,*) "Pivot vector"
   write(outunit,*) patch%ipiv
endif
if (associated(patch%x)) then
   write(outunit,*) "Solution x"
   write(outunit,*)
   do i=1,patch%n
      write(outunit,*) patch%x(i,:)
   end do
   write(outunit,*)
endif
write(outunit,*) "-----------  end patch linear system -------------"
write(outunit,*)
end subroutine print_patch_linear_system

!====================================================
! Private routines
!====================================================

!          -------------
subroutine more_vertices(patch)
!          -------------

!----------------------------------------------------
! This routine doubles the allocation for vertices in patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), target :: patch ! The patch
!----------------------------------------------------
! Local variables:

type(vertex_type), pointer :: temp(:)
integer :: oldsize, newsize, astat
!----------------------------------------------------
! Begin executable code

oldsize = size(patch%vertex)
newsize = 2*oldsize

temp => patch%vertex
allocate(patch%vertex(newsize),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in more_vertices")
   stop
endif
patch%vertex(1:oldsize) = temp
deallocate(temp)

end subroutine more_vertices

!          ----------
subroutine more_edges(patch)
!          ----------

!----------------------------------------------------
! This routine doubles the allocation for edges in patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), target :: patch ! The patch
!----------------------------------------------------
! Local variables:

type(edge_type), pointer :: temp(:)
integer :: oldsize, newsize, astat
!----------------------------------------------------
! Begin executable code

oldsize = size(patch%edge)
newsize = 2*oldsize

temp => patch%edge
allocate(patch%edge(newsize),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in more_edges")
   stop
endif
patch%edge(1:oldsize) = temp
deallocate(temp)

end subroutine more_edges

!          -------------
subroutine more_elements(patch)
!          -------------

!----------------------------------------------------
! This routine doubles the allocation for elements in patch
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), target :: patch ! The patch
!----------------------------------------------------
! Local variables:

type(element_type), pointer :: temp(:)
integer :: oldsize, newsize, astat
!----------------------------------------------------
! Begin executable code

oldsize = size(patch%element)
newsize = 2*oldsize

temp => patch%element
allocate(patch%element(newsize),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in more_elements")
   stop
endif
patch%element(1:oldsize) = temp
deallocate(temp)

end subroutine more_elements

!          -------------------
subroutine reallocate_solution(patch,elem,newelemdeg,newedgedeg)
!          -------------------

!----------------------------------------------------
! This subroutine reallocates the solution and grid_solution components
! of element elem and it's edges, and copies the existing solution, if the
! current allocation is not large enough for the given degrees.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: elem              ! The element to refine
integer, intent(in) :: newelemdeg        ! The new element degree
integer, intent(in) :: newedgedeg(EDGES_PER_ELEMENT) ! The new edge degrees
!----------------------------------------------------
! Local variables:

integer :: i, newsize, oldsize, edge, astat
real(my_real), pointer :: temp(:,:,:), temp2(:,:,:)
!----------------------------------------------------
! Begin executable code

! Check element solution allocation

newsize = element_dof(newelemdeg)
if (newsize > 0) then
   if (associated(patch%element(elem)%solution)) then
      oldsize = size(patch%element(elem)%solution,dim=1)
   else
      oldsize = 0
   endif
   if (oldsize < newsize) then
      if (oldsize /= 0) then
         temp => patch%element(elem)%solution
         nullify(patch%element(elem)%solution)
         temp2 => patch%element(elem)%grid_solution
         nullify(patch%element(elem)%grid_solution)
      endif
      allocate(patch%element(elem)%solution(newsize,patch%system_size, &
                                            patch%num_evec), &
               patch%element(elem)%grid_solution(newsize,patch%system_size, &
                                            patch%num_evec), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in patch reallocate_solution")
         stop
      endif
      if (oldsize == 0) then
         patch%element(elem)%solution = 0
         patch%element(elem)%grid_solution = 0
      else
         patch%element(elem)%solution(1:oldsize,:,:) = temp
         deallocate(temp)
         patch%element(elem)%solution(oldsize+1:newsize,:,:) = 0
         patch%element(elem)%grid_solution(1:oldsize,:,:) = temp2
         deallocate(temp2)
         patch%element(elem)%grid_solution(oldsize+1:newsize,:,:) = 0
      endif
   endif
endif

! For each edge, check the allocation of solution

do i=1,EDGES_PER_ELEMENT
   edge = patch%element(elem)%edge(i)
   newsize = newedgedeg(i) - 1
   if (associated(patch%edge(edge)%solution)) then
      oldsize = size(patch%edge(edge)%solution,dim=1)
   else
      oldsize = 0
   endif
   if (oldsize < newsize) then
      if (oldsize /= 0) then
         temp => patch%edge(edge)%solution
         nullify(patch%edge(edge)%solution)
         temp2 => patch%edge(edge)%grid_solution
         nullify(patch%edge(edge)%grid_solution)
      endif
      allocate(patch%edge(edge)%solution(newsize,patch%system_size, &
                                         patch%num_evec), &
               patch%edge(edge)%grid_solution(newsize,patch%system_size, &
                                              patch%num_evec), &
               stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in patch reallocate_solution")
         stop
      endif
      if (oldsize == 0) then
         patch%edge(edge)%solution = 0
         patch%edge(edge)%grid_solution = 0
      else
         patch%edge(edge)%solution(1:oldsize,:,:) = temp
         deallocate(temp)
         patch%edge(edge)%solution(oldsize+1:newsize,:,:) = 0
         patch%edge(edge)%grid_solution(1:oldsize,:,:) = temp2
         deallocate(temp2)
         patch%edge(edge)%grid_solution(oldsize+1:newsize,:,:) = 0
      endif
   endif
end do

end subroutine reallocate_solution

!          -----------------------------
subroutine set_patch_boundary_conditions(patch,bc)
!          -----------------------------

!----------------------------------------------------
! This routine sets the boundary conditions for the patch, i.e. the vertex
! and edge type and bmark.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: bc                ! How to set vertex type, edge type,
                                         ! and bmark for the patch boundary:
                                         ! 0 - keep the values from grid
                                         ! 1 - set all Dirichlet with
                                         !     bmark = huge(0)
                                         ! 2 - set grid interior Dirichlet with
                                         !     bmark = huge(0)
                                         ! 3 - set all Natural with 
                                         !     bmark = huge(0)
                                         ! 4 - set grid interior Natural with
                                         !     bmark = huge(0)
                                         ! 5 - set all vertices Dirichlet with
                                         !     bmark = huge(0) and interior
                                         !     edges Dirichlet bmark = huge(0)
                                         ! 6 - set all vertices Dirichlet with
                                         !     bmark = huge(0) and interior
                                         !     edges and high order edges
                                         !     Dirichlet with bmark = huge(0)
!----------------------------------------------------
! Local variables:

logical :: visited_edge(patch%nedge), visited_vert(patch%nvert)
integer :: elem, edge, vert, i, j
!----------------------------------------------------
! Begin executable code

! If bc == 0, i.e. keep the values from grid, do nothing

if (bc == 0) return

! To determine which entities are boundary, pass through the patch elements
! looking for neighbor == BOUNDARY.  If it is the first time to see the
! corresponding edge, set its boundary condition and the boundary conditions
! of its vertices that have not been set yet.

visited_edge = .false.
visited_vert = .false.

do elem=1,patch%nelem
   do i=1,NEIGHBORS_PER_ELEMENT
      if (patch%element(elem)%neighbor(i) == BOUNDARY) then
         edge = patch%element(elem)%edge(i)
         if (visited_edge(edge)) cycle
         visited_edge(edge) = .true.
         select case(bc)
         case(1)
            patch%edge(edge)%type = DIRICHLET
            patch%edge(edge)%bmark = huge(0)
         case(2)
            if (patch%edge(edge)%type(1) == INTERIOR) then
               patch%edge(edge)%type = DIRICHLET
               patch%edge(edge)%bmark = huge(0)
            endif
         case(3)
            patch%edge(edge)%type = NATURAL
            patch%edge(edge)%bmark = huge(0)
         case(4)
            if (patch%edge(edge)%type(1) == INTERIOR) then
               patch%edge(edge)%type = NATURAL
               patch%edge(edge)%bmark = huge(0)
            endif
         case(5)
            if (patch%edge(edge)%type(1) == INTERIOR) then
               patch%edge(edge)%type = DIRICHLET
               patch%edge(edge)%bmark = huge(0)
            endif
         case(6)
            if (patch%edge(edge)%type(1) == INTERIOR .or. &
                patch%edge(edge)%degree > 1) then
               patch%edge(edge)%type = DIRICHLET
               patch%edge(edge)%bmark = huge(0)
            endif
         case default
            ierr = PHAML_INTERNAL_ERROR
            call fatal("set_patch_boundary_conditions: unrecognized bc")
            stop
         end select
         do j=1,2
            vert = patch%edge(edge)%vertex(j)
            if (visited_vert(vert)) cycle
            visited_vert(vert) = .true.
            select case(bc)
            case(1)
               patch%vertex(vert)%type = DIRICHLET
               patch%vertex(vert)%bmark = huge(0)
            case(2)
               if (patch%vertex(vert)%type(1) == INTERIOR) then
                  patch%vertex(vert)%type = DIRICHLET
                  patch%vertex(vert)%bmark = huge(0)
               endif
            case(3)
               patch%vertex(vert)%type = NATURAL
               patch%vertex(vert)%bmark = huge(0)
            case(4)
               if (patch%vertex(vert)%type(1) == INTERIOR) then
                  patch%vertex(vert)%type = NATURAL
                  patch%vertex(vert)%bmark = huge(0)
               endif
            case(5)
               patch%vertex(vert)%type = DIRICHLET
               patch%vertex(vert)%bmark = huge(0)
            case(6)
               patch%vertex(vert)%type = DIRICHLET
               patch%vertex(vert)%bmark = huge(0)
            end select
         end do
      end if
   end do
end do

end subroutine set_patch_boundary_conditions

!          ------------
subroutine set_first_eq(patch,neq,n)
!          ------------

!----------------------------------------------------
! This routine determines the first patch linear system equation associated with
! each patch grid entity.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(out) :: neq              ! Return the number of equations
integer, intent(out) :: n                ! Return number of non-Dirichlet rows
!----------------------------------------------------
! Local variables:

integer :: ss, i, dof
!----------------------------------------------------
! Begin executable code

if (patch%type_of_patch == LOCAL_H_ERRIND_PATCH) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("A local h error indicator patch should not call set_first_eq")
   stop
endif

! Convenience variables

ss = patch%system_size

! First equations are the vertices.  Each vertex gets system_size equations.

neq = 0
n   = 0
do i=1,patch%nvert
   patch%vertex(i)%first_eq = neq + 1
   patch%vertex(i)%neq = ss
   neq = neq + patch%vertex(i)%neq
   n = n + count(patch%vertex(i)%type /= DIRICHLET)
end do

! Next are equations for the edges.  There are (degree-1)*system_size equations
! blocked such that the system_size equations of the same degree are together.
! first_eq == 0 indicates there are no equations for the edge.

do i=1,patch%nedge
   if (patch%edge(i)%isleaf .and. patch%edge(i)%degree > 1) then
      patch%edge(i)%first_eq = neq + 1
      patch%edge(i)%neq = ss*(patch%edge(i)%degree-1)
      neq = neq + patch%edge(i)%neq
      n = n + count(patch%edge(i)%type /= DIRICHLET)*(patch%edge(i)%degree-1)
   else
      patch%edge(i)%first_eq = 0
      patch%edge(i)%neq = 0
   endif
end do

! Finally the equations for the elements.

do i=1,patch%nelem
   if (patch%element(i)%isleaf) then
      dof = element_dof(patch%element(i)%degree)
      if (dof > 0) then
         patch%element(i)%first_eq = neq + 1
         patch%element(i)%neq = ss*dof
         neq = neq + patch%element(i)%neq
         n = n + patch%element(i)%neq
      else
         patch%element(i)%first_eq = 0
         patch%element(i)%neq = 0
      endif
   else
      patch%element(i)%first_eq = 0
      patch%element(i)%neq = 0
   endif
end do

end subroutine set_first_eq

!          ----------------
subroutine patch_edge_exact(patch,edge,sysrank)
!          ----------------

!----------------------------------------------------
! This is a modification of edge_exact in grid_util for patches and just
! setting Dirichlet boundary conditions.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: edge              ! Which edge to set
integer, intent(in) :: sysrank           ! Which component to set
!----------------------------------------------------
! Local variables:

integer :: i, j, nqp, nev, jerr, deg(4), astat, n, ss
real(my_real) :: xv(VERTICES_PER_ELEMENT),yv(VERTICES_PER_ELEMENT)
real(my_real), pointer :: weight(:), xq(:), yq(:)
real(my_real), allocatable :: basis(:,:), a(:), b(:,:), true(:,:)
integer, allocatable :: ipiv(:)
real(my_real) :: c(patch%system_size,patch%system_size),rs(patch%system_size)
integer :: itype(patch%system_size)

!----------------------------------------------------
! Begin executable code

! nothing to do if only linear bases on this edge

if (patch%edge(edge)%degree < 2) return

! nothing to do if this is not a Dirichlet edge

if (patch%edge(edge)%type(sysrank) /= DIRICHLET) return

! set special Dirichlet boundary to 0

if (patch%edge(edge)%bmark == huge(0)) then
   patch%edge(edge)%solution(:,sysrank,:) = 0
   return
endif

! useful constants

nev = patch%num_evec
ss = patch%system_size

! local copy of the vertex coordinates, plus a fake third vertex to make
! a triangle

xv(1) = patch%vertex(patch%edge(edge)%vertex(1))%coord%x
xv(2) = patch%vertex(patch%edge(edge)%vertex(2))%coord%x
xv(3) = (xv(1)+xv(2))/2
yv(1) = patch%vertex(patch%edge(edge)%vertex(1))%coord%y
yv(2) = patch%vertex(patch%edge(edge)%vertex(2))%coord%y
yv(3) = (yv(1)+yv(2))/2
if (abs(xv(2)-xv(1)) > abs(yv(2)-yv(1))) then
   yv(3) = yv(3) + 1
else
   xv(3) = xv(3) + 1
endif

! get the quadrature rule for the edge

call quadrature_rule_line(min(MAX_QUAD_ORDER_LINE,patch%edge(edge)%degree+1), &
                          xv(1:2), yv(1:2), nqp, weight, xq, yq, jerr)
if (jerr /= 0) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("Error getting quadrature rule in patch_edge_exact", &
              intlist=(/jerr/))
   stop
endif

! evaluate basis functions at the quadrature points

allocate(basis(3+patch%edge(edge)%degree-1,nqp),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in patch_edge_exact")
   stop
endif

deg = 1
deg(3) = patch%edge(edge)%degree

call p_hier_basis_func(xq,yq,xv,yv,deg,"a",basis)

! evaluate the function at the quadrature points

allocate(true(nqp,nev),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in patch_edge_exact")
   stop
endif

do i=1,nqp
   call my_bconds(xq(i),yq(i),0.0_my_real,patch%edge(edge)%bmark,itype,c,rs)
   true(i,:) = rs(sysrank)
end do

! take off the linear bases part of the function

do j=1,nev
 true(:,j) = true(:,j) - &
    patch%vertex(patch%edge(edge)%vertex(1))%solution(sysrank,j)*basis(1,:) - &
    patch%vertex(patch%edge(edge)%vertex(2))%solution(sysrank,j)*basis(2,:)
end do

! set up least squares linear system, storing the matrix in LAPACK symmetric
! packed form

n = patch%edge(edge)%degree - 1

allocate(a((n*(n+1))/2),b(n,nev),ipiv(n),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in patch_edge_exact")
   stop
endif

do i=1,n
   do j=i,n
      a(i+(j*(j-1))/2) = sum(weight*basis(3+i,:)*basis(3+j,:))
   end do
   do j=1,nev
      b(i,j) = sum(weight*basis(3+i,:)*true(:,j))
   end do
end do

! solve the least squares system

if (my_real == kind(1.0e0)) then
   call sspsv("U",n,nev,a,ipiv,b,n,jerr)
elseif (my_real == kind(1.0d0)) then
   call dspsv("U",n,nev,a,ipiv,b,n,jerr)
else
   ierr = PHAML_INTERNAL_ERROR
   call fatal("patch_edge_exact: my_real is neither single nor double precision. Can't call LAPACK")
   stop 
endif

! copy the solution to the edge data structure

do j=1,nev
   do i=1,n
      patch%edge(edge)%solution(i,sysrank,j) = b(i,j)
   end do
end do

deallocate(weight,xq,yq,basis,a,b,true,ipiv)

end subroutine patch_edge_exact

!          -----------------------
subroutine linsys_solution_to_grid(patch)
!          -----------------------

!----------------------------------------------------
! This routine copies the linear system solution to the grid solution.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
!----------------------------------------------------
! Local variables:

integer :: i, j, k, row, ss
real(my_real) :: solution(patch%neq,patch%num_evec)
!----------------------------------------------------
! Begin executable code

! A local h error indicator patch cannot be copied to the grid because
! it uses the h-hierarchical basis

if (patch%type_of_patch == LOCAL_H_ERRIND_PATCH) return

! Convenience variables

ss = patch%system_size

! The linear system solution, x, contains only the non-Dirichlet rows.  Copy
! it to a vector with the indexing of the full linear system.

solution(patch%nondir_rows,:) = patch%x

! Copy the non-Dirichlet vertex equation solutions to the grid.  The Dirichlet
! vertices already have the solution set when setting up boundary conditions.

do i=1,patch%nvert
   row = patch%vertex(i)%first_eq-1
   do j=1,patch%vertex(i)%neq
      row = row + 1
      if (.not. patch%dirichlet(row)) patch%vertex(i)%solution(j,:) = &
                                      solution(row,:)
   end do
end do

! Copy the non-Dirichlet edge equation solutions to the grid.  The Dirichlet
! edges already have the solution set when setting up boundary conditions.

do i=1,patch%nedge
   if (patch%edge(i)%first_eq == 0) cycle
   row = patch%edge(i)%first_eq-1
   do j=1,patch%edge(i)%neq/ss
      do k=1,ss
         row = row + 1
         if (.not. patch%dirichlet(row)) patch%edge(i)%solution(j,k,:) = &
                                         solution(row,:)
      end do
   end do
end do

! Copy the element equation solutions to the grid.

do i=1,patch%nelem
   if (patch%element(i)%first_eq == 0) cycle
   row = patch%element(i)%first_eq-ss
   do j=1,patch%element(i)%neq/ss
      row = row + ss
      patch%element(i)%solution(j,:,:) = solution(row:row+ss-1,:)
   end do
end do

end subroutine linsys_solution_to_grid

!          --------
subroutine assemble(patch,elem,lmat,lrhs,just_rhs,extra_rhs)
!          --------

!----------------------------------------------------
! This routine assembles elemental matrix lmat and lrhs into the patch
! linear system.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(patch_type), intent(inout) :: patch ! The patch
integer, intent(in) :: elem              ! The element in the patch
real(my_real), intent(in) :: lmat(:,:)   ! Elemental matrix
real(my_real), intent(in) :: lrhs(:)     ! Elemental right hand side
logical, intent(in) :: just_rhs          ! True is just a right hand side update
real(my_real), intent(in), optional :: extra_rhs(:,:)! For multiple eigenvectors
!----------------------------------------------------
! Local variables:

integer :: index_set(size(lrhs))
integer :: ss, neq, i, j
!----------------------------------------------------
! Begin executable code

! Convenience variables

ss = patch%system_size

! Set up a vector of indices into the patch linear system to use as a vector
! subscript into lmat and lrhs

neq = 0
do i=1,VERTICES_PER_ELEMENT
   do j=patch%vertex(patch%element(elem)%vertex(i))%first_eq, &
        patch%vertex(patch%element(elem)%vertex(i))%first_eq + &
        patch%vertex(patch%element(elem)%vertex(i))%neq - 1
      neq = neq + 1
      index_set(neq) = j
   end do
end do
do i=1,EDGES_PER_ELEMENT
   do j=patch%edge(patch%element(elem)%edge(i))%first_eq, &
        patch%edge(patch%element(elem)%edge(i))%first_eq + &
        patch%edge(patch%element(elem)%edge(i))%neq - 1
      neq = neq + 1
      index_set(neq) = j
   end do
end do
do j=patch%element(elem)%first_eq, &
     patch%element(elem)%first_eq + patch%element(elem)%neq - 1
   neq = neq + 1
   index_set(neq) = j
end do

! Add the elemental matrix and rhs into the patch linear system

patch%A(index_set(1:neq),index_set(1:neq)) = &
   patch%A(index_set(1:neq),index_set(1:neq)) + lmat(1:neq,1:neq)
patch%b(index_set(1:neq),1) = patch%b(index_set(1:neq),1) + lrhs(1:neq)
if (present(extra_rhs)) then
   patch%b(index_set(1:neq),2:patch%num_evec) = &
      patch%b(index_set(1:neq),2:patch%num_evec) + extra_rhs(1:neq,:)
endif

end subroutine assemble

end module patch_problem
