!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Mathematical and Computational Sciences Division                !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module search_tree_mod

!----------------------------------------------------
! This module implements a search tree.  It is a binary tree in which each node
! has an integer id and an integer data.  The left subtree beneath a node has
! id's smaller than node's and the right subtree has larger id's.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

!----------------------------------------------------

implicit none
private
public search_tree_type, create_search_tree, destroy_search_tree, &
       insert_search_tree, id_is_in_search_tree, get_search_tree_data, &
       ST_NO_DATA

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

integer, parameter :: ST_NO_DATA = huge(0)

!----------------------------------------------------
! The following variables are defined:

!----------------------------------------------------
! The following types are defined:

type search_tree_type
   private
   integer :: id, data
   type(search_tree_type), pointer :: lchild, rchild
end type search_tree_type

!----------------------------------------------------

contains

!          ------------------
subroutine create_search_tree(st)
!          ------------------

!----------------------------------------------------
! This routine initializes st to be empty
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(search_tree_type), intent(out) :: st
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

st%id = ST_NO_DATA
st%data = ST_NO_DATA
nullify(st%lchild,st%rchild)

end subroutine create_search_tree

!                    -------------------
recursive subroutine destroy_search_tree(st)
!                    -------------------

!----------------------------------------------------
! This routine frees all memory in st
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(search_tree_type), intent(inout) :: st
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (associated(st%lchild)) then
   call destroy_search_tree(st%lchild)
   deallocate(st%lchild)
endif

if (associated(st%rchild)) then
   call destroy_search_tree(st%rchild)
   deallocate(st%rchild)
endif

st%id = ST_NO_DATA
st%data = ST_NO_DATA

end subroutine destroy_search_tree

!                    ------------------
recursive subroutine insert_search_tree(st,id,data,errcode)
!                    ------------------

!----------------------------------------------------
! This routine inserts (id,data) into the search tree.
! errcode = 0 success
!           1 id is already in the search tree; no change
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(search_tree_type), intent(inout) :: st
integer, intent(in) :: id, data
integer, intent(out) :: errcode
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

errcode = 0

! the first entry in the tree

if (st%id == ST_NO_DATA) then
   st%id = id
   st%data = data

! found id is already in the tree

elseif (st%id == id) then
   errcode = 1

! data goes in left subtree

elseif (id < st%id) then
   if (associated(st%lchild)) then
      call insert_search_tree(st%lchild,id,data,errcode)
   else
      allocate(st%lchild)
      st%lchild%id = id
      st%lchild%data = data
      nullify(st%lchild%lchild,st%lchild%rchild)
   endif

! data goes in right subtree

else
   if (associated(st%rchild)) then
      call insert_search_tree(st%rchild,id,data,errcode)
   else
      allocate(st%rchild)
      st%rchild%id = id
      st%rchild%data = data
      nullify(st%rchild%lchild,st%rchild%rchild)
   endif

endif

end subroutine insert_search_tree

!                  --------------------
recursive function id_is_in_search_tree(st,id) result(it_is)
!                  --------------------

!----------------------------------------------------
! This routine returns true if id is in st, false if not
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(search_tree_type), intent(in) :: st
integer, intent(in) :: id
logical :: it_is
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (st%id == id) then
   it_is = .true.
elseif (id < st%id) then
   if (associated(st%lchild)) then
      it_is = id_is_in_search_tree(st%lchild,id)
   else
      it_is = .false.
   endif
else
   if (associated(st%rchild)) then
      it_is = id_is_in_search_tree(st%rchild,id)
   else
      it_is = .false.
   endif
endif

end function id_is_in_search_tree

!                  --------------------
recursive function get_search_tree_data(st,id) result(data)
!                  --------------------

!----------------------------------------------------
! This routine returns the data associated with id in st.  If id is not
! in st, then it returns ST_NO_DATA.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(search_tree_type), intent(in) :: st
integer, intent(in) :: id
integer :: data
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

if (st%id == id) then
   data = st%data
elseif (id < st%id) then
   if (associated(st%lchild)) then
      data = get_search_tree_data(st%lchild,id)
   else
      data = ST_NO_DATA
   endif
else
   if (associated(st%rchild)) then
      data = get_search_tree_data(st%rchild,id)
   else
      data = ST_NO_DATA
   endif
endif

end function get_search_tree_data

end module search_tree_mod
